import logging
from collections import OrderedDict, defaultdict
from functools import reduce
from typing import Any, Iterable, List, Optional, Tuple, Union

import numpy
import yaml
from kestrel.exceptions import IncompleteDataMapping
from kestrel.ir.filter import ReferenceValue
from kestrel.mapping.transformers import run_transformer, run_transformer_on_series
from kestrel.utils import list_folder_files
from pandas import DataFrame, Int64Dtype
from typeguard import typechecked

_logger = logging.getLogger(__name__)


PROMOTE_LIST_OP = {
    "=": "IN",
    "!=": "NOT IN",
}


def _reverse_value_map(obj: dict):
    result = defaultdict(list)
    for k, v in obj.items():
        if isinstance(v, list):
            for i in v:
                result[i].append(k)
        else:
            result[v].append(k)
    return result


def _add_mapping(obj: dict, key: str, mapping: dict):
    """Add `key` -> `mapping` to `obj`, appending if necessary"""
    existing_mapping = obj.get(key)
    if existing_mapping:
        if isinstance(existing_mapping, str):
            existing_mapping = [{"ocsf_field": existing_mapping}]
        elif isinstance(existing_mapping, dict):
            existing_mapping = [existing_mapping]
    else:
        existing_mapping = []
    native_value = mapping.get("native_value")
    if "ocsf_value" not in mapping and isinstance(native_value, dict):
        # There's a dict instead of a transformer function; reverse that
        mapping["ocsf_value"] = _reverse_value_map(native_value)
    existing_mapping.append(mapping)
    obj[key] = existing_mapping


def _reverse_dict(obj: dict, k: str, v: dict):
    """Reverse a single OCSF -> native mapping and add it to `obj`"""
    key = v["native_field"]
    mapping = {i: j for i, j in v.items() if i != "native_field"}
    mapping["ocsf_field"] = k
    _add_mapping(obj, key, mapping)


def _add_attr(obj: dict, key: str, value: str):
    """Add `key` -> `value` to `obj`, appending if necessary"""
    if key not in obj:
        obj[key] = value
    else:
        existing = obj[key]
        if isinstance(existing, str) and existing != value:
            obj[key] = [existing, value]
        elif value not in existing:
            existing.append(value)


def reverse_mapping(obj: dict, prefix: str = None, result: dict = None) -> dict:
    """Reverse the mapping of `obj`

    Newly loaded mapping from disk is OCSF -> native mapping. This function
    takes in such mapping, and reverse it to native -> OCSF mapping, which can
    be used by the frontend. The result mapping is flattened!

    To call the function: `reverse_mapping(ocsf_to_native_mapping)`

    Parameters:
        obj: mapping loaded from disk (OCSF -> native)
        prefix: key path to `obj`; used by the recursive function itself
        result: intermediate result mapping; used by the recursive function itself

    Returns:
        native -> OCSF mapping
    """
    if result is None:
        result = {}
    for k, v in obj.items():
        k = ".".join((prefix, k)) if prefix else k
        # Recurse if necessary
        if isinstance(v, str):
            _add_attr(result, v, k)
        elif isinstance(v, list):
            # Need to handle multiple mappings
            for i in v:
                if isinstance(i, str):
                    _add_attr(result, i, k)
                elif "native_field" in i:
                    _reverse_dict(result, k, i)
                else:
                    # Need to "deep" merge with current results
                    reverse_mapping(i, k, result)
        elif isinstance(v, dict):
            # First determine if this is a complex mapping or just another level
            if "native_field" in v:
                _reverse_dict(result, k, v)
            else:
                # Need to "deep" merge with current results
                reverse_mapping(v, k, result)

    return result


def _get_map_triple(d: dict, prefix: str, op: str, value) -> tuple:
    mapped_op = d.get(f"{prefix}_op")
    transform = d.get(f"{prefix}_value")
    _logger.debug("transform = %s (%s)", type(transform), transform)
    if transform:
        if isinstance(transform, dict):
            new_value = transform.get(value, value)  # FIXME
        else:
            new_value = run_transformer(transform, value)
    else:
        new_value = value
    if mapped_op:
        new_op = mapped_op
    elif isinstance(new_value, list):
        # Do we need to change the op here?
        new_op = PROMOTE_LIST_OP.get(op, op)
    else:
        new_op = op
    return (d[f"{prefix}_field"], new_op, new_value)


@typechecked
def translate_comparison_to_native(
    to_native_nested_map: dict, field: str, op: str, value: Any
) -> list:
    """Translate the (`field`, `op`, `value`) triple using data model map

    This function may be used in datasource interfaces to translate a comparison
    in the OCSF data model to the native data model, according to the data model
    mapping in `to_native_nested_map`.

    This function translates the (`field`, `op`, `value`) triple into a list of
    translated triples based on the provided data model map. The data model map
    is a dictionary that maps fields from one data model to another. For
    example, if you have a field named "user.name" in your data model, but the
    corresponding field in the native data model is "username", then you can use
    the data model map to translate the field name.

    Parameters:
        to_native_nested_map: OCSF to native nested mapping (directly from the YAML)
        field: The field name to be translated.
        op: The comparison operator.
        value: The value to be compared against.

    Returns:
        A list of translated triples.

    Raises:
        KeyError: If the field cannot be found in the data model map.
    """
    _logger.debug("comp_to_native: %s %s %s", field, op, value)
    result = []
    try:
        node = reduce(dict.__getitem__, field.split("."), to_native_nested_map)
        if isinstance(node, list):
            for i in node:
                if isinstance(i, dict):
                    result.append(_get_map_triple(i, "native", op, value))
                else:
                    result.append((i, op, value))
        elif isinstance(node, dict):
            result.append(_get_map_triple(node, "native", op, value))
        elif isinstance(node, str):
            result.append((node, op, value))
    except KeyError:
        # Pass-through
        _logger.debug(f"no mapping for {field}, pass through")
        result.append((field, op, value))
    _logger.debug("comp_to_native: return %s", result)
    return result


@typechecked
def translate_comparison_to_ocsf(
    to_ocsf_flattened_field_map: dict,
    field: str,
    op: str,
    value: Union[str, int, float, List[str], List[int], ReferenceValue],
) -> list:
    """Translate the (`field`, `op`, `value`) triple

    This function is used in the frontend to translate a comparison in
    the STIX (or, in the future, ECS) data model to the OCSF data
    model, according to the data model mapping in `to_ocsf_flattened_field_map`.

    This function translates the (`field`, `op`, `value`) triple into a list of
    translated triples based on the provided data model map. The data model map
    is a dictionary that maps fields from one data model to another. For
    example, if you have a field named "user.name" in your data model, but the
    corresponding field in the native data model is "username", then you can use
    the data model map to translate the field name.

    Parameters:
        to_ocsf_flattened_field_map: a flattened native-to-ocsf field mapping
        field: The field name to be translated.
        op: The comparison operator.
        value: The value to be compared against.

    Returns:
        A list of translated triples.

    Raises:
        KeyError: If the field cannot be found in the data model map.

    """
    _logger.debug("comp_to_ocsf: %s %s %s", field, op, value)
    result = []
    mapping = to_ocsf_flattened_field_map.get(field)
    if isinstance(mapping, str):
        # Simple 1:1 field name mapping
        result.append((mapping, op, value))
    elif isinstance(mapping, list):
        for i in mapping:
            if isinstance(i, dict):
                result.append(_get_map_triple(i, "ocsf", op, value))
            else:
                result.append((i, op, value))
    return result


@typechecked
def load_default_mapping(
    data_model_name: Optional[str],
    mapping_pkg: str = "kestrel.mapping",
    submodule: str = "fields",
):
    result = {}
    for f in list_folder_files(mapping_pkg, submodule, data_model_name, "yaml"):
        with open(f, "r") as fp:
            result.update(yaml.safe_load(fp))
    return result


@typechecked
def get_all_entity_types() -> List[str]:
    return list(set(load_default_mapping(None, submodule="types").values()))


@typechecked
def check_entity_identifier_existence_in_mapping(
    data_model_mapping: dict,
    entity_identifiers: dict,
    interface_information: Optional[str] = None,
):
    for entity_name, ids in entity_identifiers.items():
        if entity_name in data_model_mapping:
            entity = data_model_mapping[entity_name]
            for idx in ids:
                try:
                    reduce(dict.__getitem__, idx.split("."), entity)
                except KeyError:
                    msg_body = f"Identifier '{idx}' for entity '{entity_name}' is missing in data mapping"
                    appendix = (
                        f" at '{interface_information}'"
                        if interface_information
                        else ""
                    )
                    raise IncompleteDataMapping(msg_body + appendix)


@typechecked
def _get_from_mapping(mapping: Union[str, list, dict], key) -> list:
    result = []
    if isinstance(mapping, list):
        for i in mapping:
            if isinstance(i, dict):
                result.append(i[key])
            else:
                result.append(i)
    elif isinstance(mapping, dict):
        result.append(mapping[key])
    elif isinstance(mapping, str):
        result.append(mapping)
    return result


@typechecked
def translate_projection_to_native(
    to_native_nested_map: dict,  # mapping for translation
    ocsf_base_field: Optional[str],  # translate this base field
    attrs: Optional[Iterable] = None,  # known attributes (in OCSF) to project
    native_table_schema: Optional[List[str]] = None,  # additional filter
) -> list:
    result = []

    if ocsf_base_field:
        try:
            base_map = reduce(
                dict.__getitem__, ocsf_base_field.split("."), to_native_nested_map
            )
        except KeyError:
            _logger.warning(f"No mapping for base projection field: {ocsf_base_field}")
            base_map = {}
    else:
        # event does not have ocsf_base_field
        base_map = to_native_nested_map

    if attrs:
        # project specified attributes
        for attr in attrs:
            try:
                mapping = reduce(dict.__getitem__, attr.split("."), base_map)
                result.extend(
                    [
                        (i, attr)
                        for i in _get_from_mapping(mapping, "native_field")
                        if native_table_schema
                        and i in native_table_schema
                        or not native_table_schema
                    ]
                )
            except KeyError:
                # TODO: think better way than pass-through, e.g., raise exception
                _logger.warning(
                    f"mapping not found for entity: '{ocsf_base_field}' and attribute: '{attr}'; treat it as no mapping needed"
                )
                result.append((attr, attr))
    else:
        # project all attributes known for the entity
        # or event if no ocsf_base_field specified
        # filter by native_table_schema

        # filter only the ocsf_base_field part of the map
        for native_field, mapping in reverse_mapping(base_map).items():
            if not native_field.endswith("*"):
                result.extend(
                    [
                        (native_field, i)
                        for i in _get_from_mapping(mapping, "ocsf_field")
                        if native_table_schema
                        and native_field in native_table_schema
                        or not native_table_schema
                    ]
                )

    # De-duplicate list while maintaining order
    final_result = list(OrderedDict.fromkeys(result))
    _logger.debug("proj_to_native: return %s", final_result)

    return final_result


@typechecked
def translate_entity_projection_to_ocsf(
    to_ocsf_flattened_field_map: dict, native_projection: str
) -> str:
    _map = to_ocsf_flattened_field_map
    if not native_projection.endswith("*"):
        native_projection += ".*"
    ocsf_projection = _map.get(native_projection, native_projection)
    if isinstance(ocsf_projection, list):
        ocsf_projection = ocsf_projection[0]
    ocsf_projection = ocsf_projection[:-2]
    return ocsf_projection


@typechecked
def translate_attributes_projection_to_ocsf(
    to_ocsf_flattened_field_map: dict,
    native_type: str,
    entity_type: str,
    attrs: Iterable[str],
) -> Tuple:
    _map = to_ocsf_flattened_field_map
    result = []
    for attr in attrs:
        mapping = _map.get(attr)
        if not mapping and native_type:  # try extend with STIX style
            mapping = _map.get(f"{native_type}:{attr}")
        if not mapping and native_type:  # try extend with ECS style
            mapping = _map.get(f"{native_type}.{attr}")
        if mapping:
            ocsf_fields = _get_from_mapping(mapping, "ocsf_field")
            if entity_type and entity_type != "event":
                # Need to restrict attributes of that entity type
                prefix = f"{entity_type}."
                ocsf_attrs = [
                    field[len(prefix) :]
                    for field in ocsf_fields
                    if field.startswith(prefix)
                ]
                result += ocsf_attrs
        else:  # not found; pass through
            result.append(attr)
    final_result = list(OrderedDict.fromkeys(result))
    return tuple(final_result)


def _get_transformer(obj: dict) -> Optional[Union[str, dict]]:
    # First look for native -> ocsf value transformer
    ocsf_value = obj.get("ocsf_value")
    if not ocsf_value:
        # Next check if there's a ocsf -> native transformer
        native_value = obj.get("native_value")
        if isinstance(native_value, dict):
            # If it's a simple value map, we can reverse it
            ocsf_value = _reverse_value_map(native_value)
    return ocsf_value


def _get_transformers(mappings: List[Union[str, dict]]) -> List[Union[str, dict]]:
    """Collect value transformers from mappings, discarding duplicates"""
    result = []  # Need a list instead of set since dicts aren't hashable
    for mapping in mappings:
        if "ocsf_value" in mapping or "native_value" in mapping:
            transformer = _get_transformer(mapping)
            if transformer and transformer not in result:
                result.append(transformer)
    return result


@typechecked
def translate_dataframe(df: DataFrame, to_native_nested_map: dict) -> DataFrame:
    # Translate results into Kestrel OCSF data model
    # The column names of df are already mapped
    for col in df.columns:
        try:
            mapping = reduce(dict.__getitem__, col.split("."), to_native_nested_map)
        except KeyError:
            _logger.debug("No mapping for %s", col)
            mapping = None
        if isinstance(mapping, dict):
            mapping = [mapping]
        if isinstance(mapping, list):
            transformer_names = _get_transformers(mapping)
            if len(transformer_names) > 0:
                if len(transformer_names) > 1:
                    raise NotImplementedError("Multiple to OCSF value transformers")
                else:
                    transformer_name = transformer_names.pop()
                    if isinstance(transformer_name, dict):
                        # Not actually a named function; it's a literal value map
                        value_map = {}
                        for k, vl in transformer_name.items():
                            if len(vl) > 1:
                                raise NotImplementedError(
                                    "Multiple to OCSF value mapping"
                                )
                            else:
                                value_map[k] = vl[0]
                        # use .apply intead of .replace to handle type correctly
                        df[col] = df[col].apply(
                            lambda x: value_map[x] if x in value_map else x
                        )
                    else:
                        s = run_transformer_on_series(
                            transformer_name, df[col].dropna()
                        )
                        df[col] = s
                        # if the series is integers, use Int64 (Nullable int) to allow NaN/NA
                        # if not, pandas will use float64 by default, which gives .0
                        if s.dtype == numpy.int64:
                            df[col] = df[col].astype(Int64Dtype())
    df = df.replace({numpy.nan: None})
    return df



import os
from uuid import uuid4

import pytest
from pandas import DataFrame, read_csv

from kestrel.cache import InMemoryCache
from kestrel.config import load_kestrel_config
from kestrel.frontend.parser import parse_kestrel_and_update_irgraph
from kestrel.ir.graph import IRGraph, IRGraphEvaluable
from kestrel.ir.instructions import Construct, SerializableDataFrame


@pytest.fixture
def process_creation_events():
    # return a two-node graph:
    #   - Construct: table from logs_ocsf_process_creation.csv
    #   - Variable es: events pointing to the Construct
    graph = IRGraph()
    parse_kestrel_and_update_irgraph("es = NEW event [ {'id': 1} ]", graph, {})
    data_node = graph.get_nodes_by_type(Construct)[0]
    test_dir = os.path.dirname(os.path.abspath(__file__))
    data_node.data = SerializableDataFrame(read_csv(os.path.join(test_dir, "logs_ocsf_process_creation.csv")))
    return graph


@pytest.fixture
def kestrel_config():
    return load_kestrel_config()


def test_inmemory_cache_set_get_del():
    c = InMemoryCache()
    idx = uuid4()
    df = DataFrame([1, 2, 3])
    c[idx] = df
    assert df.equals(c[idx])
    del c[idx]
    assert idx not in c


def test_inmemory_cache_constructor():
    ids = [uuid4() for i in range(5)]
    df = DataFrame([1, 2, 3])
    c = InMemoryCache({x:df for x in ids})
    for u in ids:
        assert df.equals(c[u])
    for u in ids:
        del c[u]
        assert u not in c


def test_eval_new_filter_disp():
    stmt = """
proclist = NEW process [ {"name": "cmd.exe", "pid": 123}
                       , {"name": "explorer.exe", "pid": 99}
                       , {"name": "firefox.exe", "pid": 201}
                       , {"name": "chrome.exe", "pid": 205}
                       ]
browsers = proclist WHERE name = 'firefox.exe' OR name = 'chrome.exe'
DISP browsers ATTR name, pid
"""
    graph = IRGraph()
    rets = parse_kestrel_and_update_irgraph(stmt, graph, {})
    graph = IRGraphEvaluable(graph)
    c = InMemoryCache()
    mapping = c.evaluate_graph(graph, c)

    # check the return is correct
    assert len(rets) == 1
    df = mapping[rets[0].id]
    assert df.to_dict("records") == [ {"name": "firefox.exe", "pid": 201}
                                    , {"name": "chrome.exe", "pid": 205}
                                    ]
    # check whether `proclist` is cached
    proclist = graph.get_variable("proclist")
    assert c[proclist.id].to_dict("records") == [ {"name": "cmd.exe", "pid": 123}
                                                , {"name": "explorer.exe", "pid": 99}
                                                , {"name": "firefox.exe", "pid": 201}
                                                , {"name": "chrome.exe", "pid": 205}
                                                ]
    # check whether `browsers` is cached
    browsers = graph.get_variable("browsers")
    assert c[browsers.id].to_dict("records") == [ {"name": "firefox.exe", "pid": 201}
                                                , {"name": "chrome.exe", "pid": 205}
                                                ]


def test_eval_filter_with_ref():
    stmt = """
proclist = NEW process [ {"name": "cmd.exe", "pid": 123}
                       , {"name": "explorer.exe", "pid": 99}
                       , {"name": "firefox.exe", "pid": 201}
                       , {"name": "chrome.exe", "pid": 205}
                       ]
browsers = proclist WHERE name = 'firefox.exe' OR name = 'chrome.exe'
specials = proclist WHERE pid IN [123, 201]
p2 = proclist WHERE pid = browsers.pid and name = specials.name
DISP p2 ATTR name, pid
"""
    graph = IRGraph()
    rets = parse_kestrel_and_update_irgraph(stmt, graph, {})
    graph = IRGraphEvaluable(graph)
    c = InMemoryCache()
    mapping = c.evaluate_graph(graph, c)

    # check the return is correct
    assert len(rets) == 1
    df = mapping[rets[0].id]
    assert df.to_dict("records") == [ {"name": "firefox.exe", "pid": 201} ]

def test_get_virtual_copy():
    stmt = """
proclist = NEW process [ {"name": "cmd.exe", "pid": 123}
                       , {"name": "explorer.exe", "pid": 99}
                       , {"name": "firefox.exe", "pid": 201}
                       , {"name": "chrome.exe", "pid": 205}
                       ]
browsers = proclist WHERE name = 'firefox.exe' OR name = 'chrome.exe'
"""
    graph = IRGraph()
    parse_kestrel_and_update_irgraph(stmt, graph, {})
    graph = IRGraphEvaluable(graph)
    c = InMemoryCache()
    mapping = c.evaluate_graph(graph, c)
    v = c.get_virtual_copy()
    new_entry = uuid4()
    v[new_entry] = DataFrame()

    # v[new_entry] does not hit c.cache
    assert len(c.cache) == 2
    assert len(v.cache) == 3 

    # the two cache_catalog are different
    assert new_entry not in c
    assert new_entry in v
    del v[new_entry]
    assert new_entry not in v
    for u in c:
        del v[u]
    assert len(v) == 0
    assert len(c) == 2


def test_eval_find_event_to_entity(process_creation_events):
    graph = process_creation_events
    stmt = "procs = FIND process RESPONDED es WHERE device.os = 'Linux' DISP procs"
    rets = parse_kestrel_and_update_irgraph(stmt, graph, {})
    graph = IRGraphEvaluable(graph)
    c = InMemoryCache()
    mapping = c.evaluate_graph(graph, c)
    assert len(rets) == 1
    df = mapping[rets[0].id]
    assert list(df.columns) == ['cmd_line', 'name', 'pid', 'uid', 'endpoint.uid', 'endpoint.name',
       'endpoint.os', 'file.name', 'file.path', 'user.uid', 'user.name',
       'user.type_id', 'parent_process.cmd_line', 'parent_process.name',
       'parent_process.pid', 'parent_process.uid',
       'parent_process.endpoint.uid', 'parent_process.endpoint.name',
       'parent_process.endpoint.os', 'user.endpoint.uid', 'user.endpoint.name',
       'user.endpoint.os', 'file.endpoint.uid', 'file.endpoint.name',
       'file.endpoint.os']
    assert df.shape[0] == 5  # WHERE clause filtered out 4 out of 9, so 5 remains


def test_eval_find_entity_to_event(process_creation_events, kestrel_config):
    graph = process_creation_events
    stmt = """
        procs = FIND process RESPONDED es WHERE device.os = 'Linux'
        eves = FIND event ORIGINATED BY procs
        DISP eves
    """
    rets = parse_kestrel_and_update_irgraph(stmt, graph, kestrel_config["entity_identifier"])
    graph = IRGraphEvaluable(graph)
    c = InMemoryCache()
    mapping = c.evaluate_graph(graph, c)
    assert len(rets) == 1
    df = mapping[rets[0].id]

    # 1. WHERE clause filtered out 4 out of 9, so 5 remains
    # 2. In the 5, 4 are not parent process of others, only the first is parent process
    # 3. There are 4 lines of logs/events that uses the parent process
    assert df.shape[0] == 4
    assert df.shape[1] == 48  # full event: the number of columns in the csv


def test_eval_find_entity_to_entity(process_creation_events, kestrel_config):
    graph = process_creation_events
    stmt = """
        procs = FIND process RESPONDED es WHERE device.os = 'Linux'
        parents = FIND process CREATED procs
        DISP parents
    """
    rets = parse_kestrel_and_update_irgraph(stmt, graph, kestrel_config["entity_identifier"])
    graph = IRGraphEvaluable(graph)
    c = InMemoryCache()
    mapping = c.evaluate_graph(graph, c)
    assert len(rets) == 1
    df = mapping[rets[0].id]

    # 1. WHERE clause filtered out 4 out of 9, so 5 remains
    # 2. The last 4 share the same parent
    # 3. So there are 2 processes returned/displayed after dedup
    assert df.shape[0] == 2
    assert list(df.columns) == ['cmd_line', 'name', 'pid', 'uid', 'endpoint.uid', 'endpoint.name', 'endpoint.os']



from collections import Counter

import networkx.utils
from pandas import DataFrame

from kestrel.cache import InMemoryCache
from kestrel.frontend.parser import parse_kestrel_and_update_irgraph
from kestrel.ir.graph import IRGraph, IRGraphSimpleQuery
from kestrel.ir.instructions import (CACHE_INTERFACE_IDENTIFIER, Construct,
                                     DataSource, Filter, Instruction,
                                     ProjectAttrs, ProjectEntity, Reference,
                                     Return, TransformingInstruction, Variable)


def test_add_get_datasource():
    g = IRGraph()
    s = g.add_datasource("stixshifter://abc")
    s.store = "xyz"

    s = g.add_datasource(DataSource("stixshifter://abc"))
    assert len(g) == 1

    s2 = DataSource("stixshifter://abcd")
    g.add_datasource(s2)
    assert len(g) == 2

    assert set(g.get_datasources()) == {s, s2}
    g.get_datasource("stixshifter", "abc") == s


def test_add_same_node():
    g = IRGraph()
    n = Instruction()
    s = g.add_node(n)
    s = g.add_node(n)
    assert len(g) == 1


def test_get_node_by_id():
    g = IRGraph()
    n = Instruction()
    s = g.add_node(n)
    assert g.get_node_by_id(n.id) == n


def test_get_nodes_by_type_and_attributes():
    g = IRGraph()
    s = g.add_datasource("stixshifter://abc")
    v1 = g.add_node(Variable("asdf", "x", "y"), s)
    v2 = g.add_node(Variable("qwer", "u", "v"), s)
    v3 = g.add_node(Variable("123", "i", "j"), s)
    ns = g.get_nodes_by_type_and_attributes(Variable, {"name": "asdf"})
    assert ns == [v1]


def test_get_returns():
    g = IRGraph()
    s = g.add_datasource("stixshifter://abc")
    g.add_return(s)
    g.add_return(s)
    g.add_return(s)
    rets = g.get_returns()
    assert len(rets) == 3
    assert [ret.sequence for ret in rets] == [0, 1, 2]
    assert len(g.get_sink_nodes()) == 3


def test_get_variables():
    g = IRGraph()
    s = g.add_datasource("stixshifter://abc")
    v = Variable("asdf", "foo", "bar")
    v1 = g.add_node(v, s)
    v2 = g.add_node(v, s)
    v3 = g.add_node(v, s)
    vs = g.get_variables()
    assert len(vs) == 1
    assert vs[0].name == "asdf"


def test_add_get_reference():
    g = IRGraph()
    s = g.add_node(DataSource("ss://ee"))
    g.add_node(Variable("asdf", "foo", "bar"), s)
    g.add_node(Reference("asdf"))
    q1 = g.add_node(Reference("qwer"))
    q2 = g.add_node(Reference("qwer"))
    g.add_node(Variable("qwer", "foo", "bar"), s)
    g.add_node(Reference("qwer"))
    assert len(g) == 4
    assert len(g.edges()) == 2

    assert q1 == q2
    assert g.get_reference("qwer") == q1
    refs = g.get_references()
    assert refs == [q1]


def test_copy_graph():
    g = IRGraph()
    s = g.add_datasource("stixshifter://abc")
    g2 = g.copy()
    assert s in g2
    for n in g2.nodes():
        n.datasource = "eee"
    assert s in g
    assert s.datasource == "eee"


def test_deepcopy_graph():
    g = IRGraph()
    s = g.add_datasource("stixshifter://abc")
    g2 = g.deepcopy()
    assert len(g2.nodes()) == 1
    s2 = list(g2.nodes())[0]
    s2.datasource = "eee"
    assert s.datasource == "abc"
    assert s2.datasource == "eee"


def test_update_graph():
    g = IRGraph()
    s = g.add_datasource("stixshifter://abc")
    v1 = g.add_node(Variable("asdf", "foo", "bar"), s)
    v2 = g.add_node(Variable("asdf", "foo", "bar"), s)
    v3 = g.add_node(Variable("asdf", "foo", "bar"), s)
    r1 = g.add_return(v3)

    g2 = IRGraph()
    s2 = g2.add_datasource("stixshifter://abc")
    v4 = g2.add_node(Variable("asdf", "foo", "bar"), g2.add_node(Reference("asdf")))
    v5 = g2.add_node(Variable("asdf", "foo", "bar"), g2.add_node(TransformingInstruction(), s2))
    r2 = g2.add_return(v5)

    assert v1.version == 0
    assert v2.version == 1
    assert v3.version == 2
    assert v4.version == 0
    assert v5.version == 1
    assert r1.sequence == 0
    assert r2.sequence == 0
    assert len(g) == 5
    assert len(g2) == 6

    g.update(g2)
    assert v1.version == 0
    assert v2.version == 1
    assert v3.version == 2
    assert v4.version == 3
    assert v5.version == 4
    assert r1.sequence == 0
    assert r2.sequence == 1
    assert len(g) == 9
    assert s2 not in g
    assert r1 in g
    assert r2 in g
    assert not g.get_references()
    assert (v3, v4) in g.edges()
    assert g.in_degree(v4) == 1
    assert g.out_degree(v4) == 0


def test_serialization_deserialization():
    g1 = IRGraph()
    s = g1.add_node(DataSource("ss://ee"))
    r = g1.add_node(Reference("asdf"))
    v = g1.add_node(Variable("asdf", "foo", "bar"), s)
    j = g1.to_json()
    g2 = IRGraph(j)
    assert s in g2.nodes()
    assert v in g2.nodes()
    assert len(g2) == 3
    assert g2.edges() == {(s,v)}


def test_find_cached_dependent_subgraph_of_node():
    g = IRGraph()

    a1 = g.add_node(DataSource("ss://ee"))
    a2 = g.add_node(Variable("asdf", "foo", "bar"), a1)
    a3 = g.add_node(Instruction())
    g.add_edge(a2, a3)
    a4 = g.add_node(Variable("qwer", "foo", "bar"), a3)

    b1 = g.add_node(DataSource("ss://eee"))
    b2 = g.add_node(Variable("asdfe", "foo", "bar"), b1)
    b3 = g.add_node(Instruction())
    g.add_edge(b2, b3)
    b4 = g.add_node(Variable("qwere", "foo", "bar"), b3)

    c1 = g.add_node(Instruction())
    g.add_edge(a4, c1)
    g.add_edge(b4, c1)
    c2 = g.add_node(Variable("zxcv", "foo", "bar"), c1)

    g2 = g.find_cached_dependent_subgraph_of_node(c2, InMemoryCache())
    assert networkx.utils.graphs_equal(g, g2)

    g3 = g.find_cached_dependent_subgraph_of_node(c2, InMemoryCache({a2.id: DataFrame(), b2.id: DataFrame()}))
    g.remove_node(a1)
    g.remove_node(b1)
    assert networkx.utils.graphs_equal(g, g3)


def test_find_dependent_subgraphs_of_node_just_cache():
    huntflow = """
p1 = NEW process [ {"name": "cmd.exe", "pid": 123}
                 , {"name": "explorer.exe", "pid": 99}
                 , {"name": "firefox.exe", "pid": 201}
                 , {"name": "chrome.exe", "pid": 205}
                 ]

browsers = p1 WHERE name = 'firefox.exe' OR name = 'chrome.exe'

DISP browsers ATTR name
"""
    graph = IRGraph()
    rets = parse_kestrel_and_update_irgraph(huntflow, graph, {})
    c = InMemoryCache()
    ret = rets[0]
    gs = graph.find_dependent_subgraphs_of_node(ret, c)
    assert len(gs) == 1
    assert len(gs[0]) == 6
    assert Counter(map(type, gs[0].nodes())) == Counter([Filter, Variable, Variable, Construct, ProjectAttrs, Return])
    assert gs[0].interface == CACHE_INTERFACE_IDENTIFIER


def test_get_trunk_n_branches_filter():
    stmt = """
x = NEW process [ {"name": "cmd.exe", "pid": 123} ]
y = NEW process [ {"asdf": "abc.exe"} ]
z = x WHERE foo = 'bar' OR baz = y.asdf
"""
    graph = IRGraph()
    parse_kestrel_and_update_irgraph(stmt, graph, {})
    trunk, r2n = graph.get_trunk_n_branches(graph.get_nodes_by_type(Filter)[0])
    assert trunk.name == "x"
    for r,n in r2n.items():
        assert next(graph.predecessors(n)).name == r.reference


def test_get_trunk_n_branches_variable():
    huntflow = """
p1 = NEW process [ {"name": "cmd.exe", "pid": 123}
                 , {"name": "explorer.exe", "pid": 99}
                 , {"name": "firefox.exe", "pid": 201}
                 , {"name": "chrome.exe", "pid": 205}
                 ]
"""
    graph = IRGraph()
    parse_kestrel_and_update_irgraph(huntflow, graph, {})
    trunk, r2n = graph.get_trunk_n_branches(graph.get_variable("p1"))
    assert isinstance(trunk, Construct)
    assert r2n == {}


def test_find_dependent_subgraphs_of_node():
    huntflow = """
p1 = NEW process [ {"name": "cmd.exe", "pid": 123}
                 , {"name": "explorer.exe", "pid": 99}
                 , {"name": "firefox.exe", "pid": 201}
                 , {"name": "chrome.exe", "pid": 205}
                 ]

browsers = p1 WHERE name = 'firefox.exe' OR name = 'chrome.exe'

p2 = GET process FROM elastic://edr1
     WHERE name = "cmd.exe"
     LAST 5 DAYS

p21 = p2 WHERE parent.name = "winword.exe"

p3 = GET process FROM stixshifter://edr2
     WHERE parent_ref.name = "powershell.exe"
     LAST 24 HOURS

p31 = p3 WHERE parent.name = "excel.exe"

p4 = p21 WHERE pid = p1.pid
p5 = GET process FROM stixshifter://edr5 WHERE pid = p4.pid

DISP p5 ATTR pid, name, cmd_line
"""
    graph = IRGraph()
    parse_kestrel_and_update_irgraph(huntflow, graph, {})

    p1 = graph.get_variable("p1")
    p2 = graph.get_variable("p2")
    p3 = graph.get_variable("p3")
    p21 = graph.get_variable("p21")
    p31 = graph.get_variable("p31")
    p4 = graph.get_variable("p4")
    p5 = graph.get_variable("p5")
    ret = graph.get_returns()[0]

    c = InMemoryCache()
    gs = graph.find_dependent_subgraphs_of_node(ret, c)
    assert len(gs) == 2
    p1_projattr = [n for n in graph.successors(p1) if isinstance(n, ProjectAttrs)][0]
    assert len(gs[0]) == 3
    assert set(map(type, gs[0].nodes())) == {Variable, ProjectAttrs, Construct}
    assert p1_projattr == gs[0].get_nodes_by_type(ProjectAttrs)[0]
    assert len(gs[1]) == 6
    assert Counter(map(type, gs[1].nodes())) == Counter([Filter, Filter, Variable, Variable, ProjectEntity, DataSource])

    c.evaluate_graph(gs[0], c)
    assert p1_projattr.id in c
    assert p1.id in c
    assert len(c) == 2
    gs = graph.find_dependent_subgraphs_of_node(ret, c)
    assert len(gs) == 1
    assert len(gs[0]) == 11
    assert p2 in gs[0]
    assert p21 in gs[0]
    assert p4 in gs[0]
    assert Counter(map(type, gs[0].nodes())) == Counter([Filter, Filter, Filter, Variable, Variable, Variable, Variable, ProjectEntity, DataSource, ProjectAttrs, ProjectAttrs])

    p4_projattr = next(graph.successors(p4))
    c[p4_projattr.id] = DataFrame()
    gs = graph.find_dependent_subgraphs_of_node(ret, c)
    assert len(gs) == 1
    assert len(gs[0]) == 8
    assert p4_projattr.id in c
    assert p4_projattr in gs[0]
    assert p5 in gs[0]
    assert ret in gs[0]
    assert Counter(map(type, gs[0].nodes())) == Counter([Filter, Return, Variable, Variable, ProjectEntity, DataSource, ProjectAttrs, ProjectAttrs])


def test_find_simple_query_subgraphs():
    huntflow = """
p1 = GET process FROM elastic://edr1
     WHERE name = "cmd.exe"
     LAST 5 DAYS

p2 = GET process FROM elastic://edr1
     WHERE pid = 999
     LAST 30 MINUTES

p3 = p1 WHERE pid = p2.pid

p4 = GET process FROM elastic://edr2 WHERE name = p3.name

DISP p4
"""
    graph = IRGraph()
    parse_kestrel_and_update_irgraph(huntflow, graph, {})
    c = InMemoryCache()
    gs = graph.find_dependent_subgraphs_of_node(graph.get_returns()[0], c)
    assert len(gs) == 1
    assert networkx.utils.graphs_equal(graph, gs[0])

    vs = set(["p1", "p2"])
    for g in gs[0].find_simple_query_subgraphs(c):
        assert isinstance(g, IRGraphSimpleQuery)
        assert Counter(map(type, g.nodes())) == Counter([Variable, Filter, ProjectEntity, DataSource])
        assert len(g.edges()) == 3
        varname = g.get_variables()[0].name
        assert varname in vs
        vs.remove(varname)
    assert vs == set()

    p1 = gs[0].get_variable("p1")
    c[p1.id] = DataFrame()
    p2 = gs[0].get_variable("p2")
    c[p2.id] = DataFrame()

    gs = graph.find_dependent_subgraphs_of_node(graph.get_returns()[0], c)
    # just a dep graph in cache
    assert len(gs) == 1
    assert Counter(map(type, gs[0].nodes())) == Counter([Variable, Variable, Filter, ProjectAttrs, ProjectAttrs, Variable])
    sinks = gs[0].get_sink_nodes()
    assert len(sinks) == 1
    sink = sinks[0]
    assert isinstance(sink, ProjectAttrs) and sink.attrs == ('name',)
    c[sink.id] = DataFrame()

    gs = graph.find_dependent_subgraphs_of_node(graph.get_returns()[0], c)
    assert len(gs) == 1
    assert sink in gs[0]
    assert Counter(map(type, gs[0].nodes())) == Counter([Variable, Filter, ProjectAttrs, DataSource, Return, ProjectEntity, Variable])
    for g in gs[0].find_simple_query_subgraphs(c):
        assert Counter(map(type, g.nodes())) == Counter([ProjectAttrs, Variable, Filter, ProjectEntity, DataSource])
        assert sink in g


def test_interface_with_two_stores():
    huntflow = """
p1 = GET process FROM elastic://edr
     WHERE name = "cmd.exe"
     LAST 5 DAYS

p2 = GET process FROM elastic://firewall
     WHERE pid = 999
     LAST 30 MINUTES

p3 = p1 WHERE pid = p2.pid

p4 = GET process FROM elastic://edr WHERE name = p3.name

DISP p4
"""
    graph = IRGraph()
    parse_kestrel_and_update_irgraph(huntflow, graph, {})
    for ds in graph.get_nodes_by_type(DataSource):
        ds.store = ds.datasource
    c = InMemoryCache()
    gs = graph.find_dependent_subgraphs_of_node(graph.get_returns()[0], c)
    assert len(gs) == 2
    assert len(gs[0]) == 4
    assert len(gs[1]) == 5



from dataclasses import dataclass
from typing import List, Mapping, Union

from mashumaro.mixins.json import DataClassJSONMixin
from pandas import DataFrame


@dataclass
class NativeQuery(DataClassJSONMixin):
    # which query language
    language: str
    # what query statement
    statement: str


@dataclass
class AnalyticOperation(DataClassJSONMixin):
    # which interface
    interface: str
    # operation description
    operation: str


@dataclass
class GraphletExplanation(DataClassJSONMixin):
    # serialized IRGraph
    graph: Mapping
    # data source query
    action: Union[NativeQuery, AnalyticOperation]


@dataclass
class GraphExplanation(DataClassJSONMixin):
    graphlets: List[GraphletExplanation]


# Kestrel Display Object
Display = Union[
    str,
    dict,
    DataFrame,
    GraphExplanation,
]



import logging
import os
from pathlib import Path
from typing import List, Mapping, Union

import pandas
import yaml
from kestrel.exceptions import (
    InvalidKestrelConfig,
    InvalidKestrelRelationTable,
    InvalidYamlInConfig,
)
from kestrel.utils import list_folder_files, load_data_file, update_nested_dict
from typeguard import typechecked

CONFIG_DIR_DEFAULT = Path.home() / ".config" / "kestrel"
CONFIG_PATH_DEFAULT = CONFIG_DIR_DEFAULT / "kestrel.yaml"
CONFIG_PATH_ENV_VAR = "KESTREL_CONFIG"  # override CONFIG_PATH_DEFAULT if provided

relations = []
entity_types = []

_logger = logging.getLogger(__name__)


@typechecked
def load_leaf_yaml(config: Mapping, path_dir: str) -> Mapping:
    new = {}
    for k, v in config.items():
        if isinstance(v, Mapping):
            new[k] = load_leaf_yaml(v, path_dir)
        elif isinstance(v, str) and v.endswith(".yaml"):
            try:
                if os.path.isabs(v):
                    with open(v, "r") as fp:
                        new[k] = yaml.safe_load(fp.read())
                else:
                    with open(os.path.join(path_dir, v), "r") as fp:
                        new[k] = yaml.safe_load(fp.read())
            except:
                raise InvalidYamlInConfig(v)
        else:
            new[k] = v
    return new


@typechecked
def load_default_config() -> Mapping:
    _logger.debug(f"Loading default config file...")
    default_config = load_data_file("kestrel.config", "kestrel.yaml")
    config_with_envvar_expanded = os.path.expandvars(default_config)
    config_content = yaml.safe_load(config_with_envvar_expanded)
    return config_content


@typechecked
def load_user_config(
    config_path_env_var: str, config_path_default: Union[str, Path]
) -> Mapping:
    config_path_default = config_path_default.absolute().as_posix()
    config_path = os.getenv(config_path_env_var, config_path_default)
    config_path = os.path.expanduser(config_path)
    config = {}
    if config_path:
        try:
            with open(config_path, "r") as fp:
                _logger.debug(f"User configuration file found: {config_path}")
                config = yaml.safe_load(os.path.expandvars(fp.read()))
            config = load_leaf_yaml(config, os.path.dirname(config_path))
        except FileNotFoundError:
            _logger.debug(f"User configuration file not exist.")
    return config


@typechecked
def load_kestrel_config() -> Mapping:
    config_default = load_default_config()
    config_user = load_user_config(CONFIG_PATH_ENV_VAR, CONFIG_PATH_DEFAULT)
    _logger.debug(f"User configuration loaded: {config_user}")
    _logger.debug(f"Updating default config with user config...")
    full_config = update_nested_dict(config_default, config_user)

    # valid the entity identifier section format
    for entity, idx in full_config["entity_identifier"].items():
        if not (isinstance(idx, list) and all((isinstance(x, str) for x in idx))):
            raise InvalidKestrelConfig(f"Invalid entity_identifier for '{entity}'")

    return full_config


@typechecked
def load_relation_configs(table_name: str) -> pandas.DataFrame:
    """Load relation tables

    Parameters:
        table_name: "entity" (entity-to-entity relation) or "event" (entity-to-event relation)

    Returns:
        Relation table in DataFrame, which has column names
    """
    filepaths = list(
        list_folder_files("kestrel.config", "relations", table_name, "csv")
    )
    if len(filepaths) > 1:
        _logger.error(f"More than one relation table found; will return the first one")
    try:
        table = pandas.read_csv(filepaths[0])
    except:
        raise InvalidKestrelRelationTable(filepaths[0])
    return table


@typechecked
def get_all_relations() -> List[str]:
    global relations
    if not relations:
        _relations = set()
        for filepath in list_folder_files(
            "kestrel.config", "relations", extension="csv"
        ):
            table = pandas.read_csv(filepath)
            _relations |= set(table["Relation"].to_list())
        relations = list(_relations)
    return relations



from datetime import datetime

import pytest
# Use sqlite3 for testing
import sqlalchemy
from dateutil import parser

from kestrel.interface.codegen.sql import SqlTranslator
from kestrel.ir.filter import (ExpOp, IntComparison, ListComparison, ListOp,
                               MultiComp, NumCompOp, StrComparison, StrCompOp,
                               TimeRange)
from kestrel.ir.instructions import Filter, Limit, Offset, ProjectAttrs, Sort


def _dt(timestr: str) -> datetime:
    return parser.parse(timestr)


def _time2string(ts: datetime) -> str:
    return ts.strftime('%Y-%m-%dT%H:%M:%S.%f')


def _remove_nl(s):
    return s.replace('\n', '')


@pytest.mark.parametrize(
    "iseq, sql", [
        # Try a simple filter
        ([Filter(IntComparison('foo', NumCompOp.GE, 0))],
         "SELECT DISTINCT * FROM my_table WHERE foo >= ?"),
        # Try a simple filter with sorting
        ([Filter(IntComparison('foo', NumCompOp.GE, 0)), Sort('bar')],
         "SELECT DISTINCT * FROM my_table WHERE foo >= ? ORDER BY bar DESC"),
        # Simple filter plus time range
        ([Filter(IntComparison('foo', NumCompOp.GE, 0), timerange=TimeRange(_dt('2023-12-06T08:17:00Z'), _dt('2023-12-07T08:17:00Z')))],
         "SELECT DISTINCT * FROM my_table WHERE foo >= ? AND timestamp >= ? AND timestamp < ?"),
        # sqlalchemy's sqlite dialect seems to always add the offset
        ([Limit(3), ProjectAttrs(['foo', 'bar', 'baz']), Filter(StrComparison('foo', StrCompOp.EQ, 'abc'))],
         "SELECT DISTINCT foo, bar, baz FROM my_table WHERE foo = ? LIMIT ? OFFSET ?"),
        # Same as above but reverse order
        ([Filter(StrComparison('foo', StrCompOp.EQ, 'abc')), ProjectAttrs(['foo', 'bar', 'baz']), Limit(3)],
         "SELECT DISTINCT foo, bar, baz FROM my_table WHERE foo = ? LIMIT ? OFFSET ?"),
        ([Filter(ListComparison('foo', ListOp.NIN, ['abc', 'def']))],
         "SELECT DISTINCT * FROM my_table WHERE (foo NOT IN (__[POSTCOMPILE_foo_1]))"), # POSTCOMPILE is some SQLAlchemy-ism
        ([Filter(StrComparison('foo', StrCompOp.MATCHES, '.*abc.*'))],
         "SELECT DISTINCT * FROM my_table WHERE foo REGEXP ?"),
        ([Filter(StrComparison('foo', StrCompOp.NMATCHES, '.*abc.*'))],
         "SELECT DISTINCT * FROM my_table WHERE foo NOT REGEXP ?"),
        ([Filter(MultiComp(ExpOp.OR, [IntComparison('foo', NumCompOp.EQ, 1), IntComparison('bar', NumCompOp.EQ, 1)]))],
         "SELECT DISTINCT * FROM my_table WHERE foo = ? OR bar = ?"),
        ([Filter(MultiComp(ExpOp.AND, [IntComparison('foo', NumCompOp.EQ, 1), IntComparison('bar', NumCompOp.EQ, 1)]))],
         "SELECT DISTINCT * FROM my_table WHERE foo = ? AND bar = ?"),
        ([Limit(1000), Offset(2000)],
         "SELECT DISTINCT * FROM my_table LIMIT ? OFFSET ?"),
    ]
)
def test_sql_translator(iseq, sql):
    trans = SqlTranslator(sqlalchemy.dialects.sqlite.dialect(), "my_table", ["foo", "bar", "baz"], None, None, _time2string, "timestamp")
    for i in iseq:
        trans.add_instruction(i)
    result = trans.result()
    assert _remove_nl(str(result)) == sql



# Lark Transformer

import logging
from datetime import datetime, timedelta, timezone
from itertools import chain
from typing import List, Union

from dateutil.parser import parse as to_datetime
from kestrel.exceptions import (
    DuplicatedRelationMapping,
    InvalidComparison,
    MissingEntityIdentifierInConfig,
    UnsupportedObjectRelation,
)
from kestrel.ir.filter import (
    BoolExp,
    ExpOp,
    FComparison,
    FExpression,
    FloatComparison,
    IntComparison,
    ListComparison,
    ListOp,
    MultiComp,
    NumCompOp,
    RefComparison,
    ReferenceValue,
    StrComparison,
    StrCompOp,
    TimeRange,
)
from kestrel.ir.graph import IRGraph
from kestrel.ir.instructions import (
    Analytic,
    AnalyticsInterface,
    Construct,
    DataSource,
    Explain,
    Filter,
    Information,
    Instruction,
    Limit,
    Offset,
    ProjectAttrs,
    ProjectEntity,
    Return,
    Sort,
    Variable,
)
from kestrel.mapping.data_model import (
    translate_attributes_projection_to_ocsf,
    translate_comparison_to_ocsf,
    translate_entity_projection_to_ocsf,
)
from kestrel.utils import unescape_quoted_string
from lark import Token, Transformer
from pandas import DataFrame
from typeguard import typechecked

_logger = logging.getLogger(__name__)


DEFAULT_VARIABLE = "_"
DEFAULT_SORT_ORDER = "DESC"


@typechecked
def _unescape_quoted_string(s: str) -> str:
    if s.startswith("r"):
        return s[2:-1]
    else:
        return s[1:-1].encode("utf-8").decode("unicode_escape")


@typechecked
def _trim_ocsf_base_field(field: str, only_trim_event: bool) -> str:
    """Remove event name (as prefix) if the field starts with an event"""
    items = field.split(".")
    if (not only_trim_event) or (
        only_trim_event
        and (
            items[0].endswith("_event")
            or items[0].endswith("_activity")
            or items[0] == "event"
        )
    ):
        items = items[1:]
    field = ".".join(items)
    return field


@typechecked
def _create_comp(
    field: str,
    op_value: str,
    value: Union[str, int, float, List, ReferenceValue],
    only_trim_event: bool,
) -> FComparison:
    # TODO: implement MultiComp

    if op_value in (ListOp.IN, ListOp.NIN):
        op = ListOp
        compf = RefComparison if isinstance(value, ReferenceValue) else ListComparison
    elif isinstance(value, int):
        op = NumCompOp
        compf = IntComparison
    elif isinstance(value, float):
        op = NumCompOp
        compf = FloatComparison
    elif isinstance(value, ReferenceValue):
        op = ListOp
        op_value = ListOp.IN if op_value in (ListOp.IN, StrCompOp.EQ) else ListOp.NIN
        compf = RefComparison
    else:
        op = StrCompOp
        compf = StrComparison

    if compf is RefComparison:
        comp = compf(
            [_trim_ocsf_base_field(field, only_trim_event)], op(op_value), value
        )
    else:
        comp = compf(_trim_ocsf_base_field(field, only_trim_event), op(op_value), value)

    return comp


@typechecked
def _map_filter_exp(
    native_projection_field: str,
    ocsf_projection_field: str,
    filter_exp: FExpression,
    field_map: dict,
    only_trim_event: bool,
) -> FExpression:
    if isinstance(
        filter_exp,
        (IntComparison, FloatComparison, StrComparison, ListComparison, RefComparison),
    ):
        # get the field/key
        if hasattr(filter_exp, "field"):
            field = filter_exp.field
        elif hasattr(filter_exp, "fields"):
            if len(filter_exp.fields) > 1:
                raise NotImplementedError(
                    "Kestrel syntax does not support fields tuple yet"
                )
            field = filter_exp.fields[0]
        else:
            raise InvalidComparison(filter_exp)

        map_result = translate_comparison_to_ocsf(
            field_map, field, filter_exp.op, filter_exp.value
        )
        # there is a case that `field` omits the return entity (prefix)
        # this is only alloed when it refers to the return entity
        # add mapping for those cases
        for full_field in (
            f"{native_projection_field}:{field}",
            f"{native_projection_field}.{field}",
        ):
            for triple in translate_comparison_to_ocsf(
                field_map, full_field, filter_exp.op, filter_exp.value
            ):
                if (
                    triple[0].startswith(ocsf_projection_field + ".")
                    and triple not in map_result
                ):
                    map_result.append(triple)

        # Build a MultiComp if field maps to several values
        if len(map_result) > 1:
            filter_exp = MultiComp(
                ExpOp.OR,
                [
                    _create_comp(field, op, value, only_trim_event)
                    for field, op, value in map_result
                ],
            )
        elif len(map_result) == 1:  # it maps to a single value
            mapping = map_result.pop()
            _logger.debug("mapping = %s", mapping)
            field = mapping[0]
            filter_exp.field = _trim_ocsf_base_field(field, only_trim_event)
            filter_exp.op = mapping[1]
            filter_exp.value = mapping[2]
        else:  # pass-through
            if field.startswith(ocsf_projection_field + "."):
                _field = field
            else:
                _field = ocsf_projection_field + "." + field
            _field = _trim_ocsf_base_field(_field, only_trim_event)
            if _field != field:
                if hasattr(filter_exp, "field"):
                    filter_exp.field = _field
                elif hasattr(filter_exp, "fields"):
                    filter_exp.fields[0] = _field

        # TODO: for RefComparison, map the attribute in value (may not be possible here)

    elif isinstance(filter_exp, BoolExp):
        # recursively map boolean expressions
        filter_exp = BoolExp(
            _map_filter_exp(
                native_projection_field,
                ocsf_projection_field,
                filter_exp.lhs,
                field_map,
                only_trim_event,
            ),
            filter_exp.op,
            _map_filter_exp(
                native_projection_field,
                ocsf_projection_field,
                filter_exp.rhs,
                field_map,
                only_trim_event,
            ),
        )
    elif isinstance(filter_exp, MultiComp):
        # normally, this should be unreachable
        # if this becomes a valid case, we need to change
        # the definition of MultiComp to accept a MultiComp
        # in addition to Comparisons in its `comps` list
        filter_exp = MultiComp(
            filter_exp.op,
            [
                _map_filter_exp(
                    native_projection_field,
                    ocsf_projection_field,
                    x,
                    field_map,
                    only_trim_event,
                )
                for x in filter_exp.comps
            ],
        )
    return filter_exp


@typechecked
def _get_entity_event_relation_projection(
    table: DataFrame,
    output_type: str,
    relation: str,
) -> str:
    t1 = table[table["OutputType"] == output_type]
    if t1.empty:
        raise UnsupportedObjectRelation("event", output_type)
    else:
        t2 = t1[t1["Relation"] == relation]
        if t2.empty:
            supported_relations = t1["Relation"].tolist()
            raise UnsupportedObjectRelation(
                "event",
                relation,
                output_type,
                f"Supported: {supported_relations}",
            )
        elif t2.shape[0] > 1:
            raise DuplicatedRelationMapping(
                "event", relation, output_type, output_projections, t2
            )
        else:
            return t2["OutputProjection"].iloc[0]


@typechecked
def _get_entity_entity_relation_specifier_projection(
    table: DataFrame,
    input_type: str,
    output_type: str,
    relation: str,
) -> (str, str):
    t1 = table[
        (table["OutputType"] == output_type) & (table["InputType"] == input_type)
    ]
    if t1.empty:
        raise UnsupportedObjectRelation(input_type, output_type)
    else:
        t2 = t1[t1["Relation"] == relation]
        if t2.empty:
            supported_relations = t1["Relation"].tolist()
            raise UnsupportedObjectRelation(
                input_type,
                relation,
                output_type,
                f"Supported: {supported_relations}",
            )
        elif t2.shape[0] > 1:
            raise DuplicatedRelationMapping(input_type, relation, output_type, t2)
        else:
            return t2["InputSpecifier"].iloc[0], t2["OutputProjection"].iloc[0]


@typechecked
def _create_filter_for_find(
    entity_identifier_map: dict,
    input_var: Variable,
    input_specifier: str,
):
    if input_var.entity_type not in entity_identifier_map:
        raise MissingEntityIdentifierInConfig(
            input_var.entity_type, entity_identifier_map
        )
    else:
        identifiers = entity_identifier_map[input_var.entity_type]
        ref_val = ReferenceValue(input_var.name, tuple(identifiers))
        comp_fields = [input_specifier + "." + x for x in identifiers]
        return Filter(RefComparison(comp_fields, ListOp.IN, ref_val))


@typechecked
class _KestrelT(Transformer):
    """Kestrel Lark Transformer

    Returns for different methods:
        - statement: [Return]
        - assignment: [Return]
        - command_no_result: [Return]
        - expression: (Instruction(root), str(entity_type), str(native_type))
        - command_with_result: (Instruction(root), str(entity_type), str(native_type))
    """

    def __init__(
        self,
        irgraph: IRGraph,
        field_map: dict,
        type_map: dict,
        entity_entity_relation_table: DataFrame,
        entity_event_relation_table: DataFrame,
        entity_identifier_map: dict,
        token_prefix: str = "",
        default_sort_order: str = DEFAULT_SORT_ORDER,
    ):
        # token_prefix is the modification by Lark when using `merge_transformers()`
        self.irgraph = irgraph
        self.default_sort_order = default_sort_order
        self.token_prefix = token_prefix
        self.type_map = type_map
        self.field_map = field_map
        self.entity_identifier_map = entity_identifier_map
        self.entity_entity_relation_table = entity_entity_relation_table
        self.entity_event_relation_table = entity_event_relation_table
        super().__init__()

    def start(self, args) -> List[Return]:
        """Parse/transform statement, and update IRGraph

        All statements should return a list of Return.
        Merge and return all of them.

        self.irgraph is updated in each statement transformer. This is required
        to search for a variable in previous IRGraph and graph generated in
        this code block.
        """
        return list(chain(*args))

    def statement(self, args) -> List[Return]:
        return args[0]

    def assignment(self, args) -> List[Return]:
        # TODO: x = y + z
        root, entity_type, native_type = args[1]
        variable = Variable(args[0].value, entity_type, native_type)
        self.irgraph.add_node(variable, root)
        return []

    def expression(self, args) -> (Instruction, str, str):
        # TODO: add more clauses than WHERE and ATTR
        # TODO: think about order of clauses when turning into nodes
        variable = args[0]
        root = variable
        if len(args) > 1:
            for clause in args[1:]:
                if isinstance(clause, Filter):
                    clause.exp = _map_filter_exp(
                        variable.native_type,
                        variable.native_type,
                        clause.exp,
                        self.field_map,
                        False,
                    )
                root = self.irgraph.add_node(clause, root)
        return root, variable.entity_type, variable.native_type

    def vtrans(self, args) -> Variable:
        if len(args) == 1:
            return self.irgraph.get_variable(args[0].value)
        else:
            # TODO: transformer support
            ...

    def new(self, args) -> (Instruction, str, str):
        native_type = None
        if len(args) == 2:
            native_type = args[0].value
            data = args[1]
        data_node = self.irgraph.add_node(Construct(data, native_type))
        if not native_type:
            raise NotImplementedError("Infer type from NEW")
        entity_type = self.type_map.get(native_type, native_type)
        return data_node, entity_type, native_type

    def var_data(self, args):
        if isinstance(args[0], Token):
            # TODO
            ...
        else:
            v = args[0]
        return v

    def json_objs(self, args):
        return args

    def json_obj(self, args):
        return dict(args)

    def json_pair(self, args):
        v = args[0].value
        if "ESCAPED_STRING" in args[0].type:
            v = unescape_quoted_string(v)
        return v, args[1]

    def json_value(self, args):
        v = args[0].value
        if args[0].type == self.token_prefix + "ESCAPED_STRING":
            v = unescape_quoted_string(v)
        elif args[0].type == self.token_prefix + "NUMBER":
            v = float(v) if "." in v else int(v)
        return v

    def variables(self, args) -> List[Variable]:
        return [self.irgraph.get_variable(arg.value) for arg in args]

    def get(self, args) -> (Instruction, str, str):
        # 0. get information of projection and return
        native_projection_field = args[0].value
        ocsf_projection_field = translate_entity_projection_to_ocsf(
            self.field_map, native_projection_field
        )
        output_type = self.type_map.get(ocsf_projection_field, ocsf_projection_field)

        # 1. process DataSource
        source_node = self.irgraph.add_node(args[1])

        # 2. process Filter
        filter_node = args[2]
        filter_node.exp = _map_filter_exp(
            native_projection_field,
            ocsf_projection_field,
            filter_node.exp,
            self.field_map,
            True,
        )
        filter_node = self.irgraph.add_node(filter_node, source_node)

        # 3. process ProjectEntity
        projection_node = self.irgraph.add_node(
            ProjectEntity(ocsf_projection_field, native_projection_field), filter_node
        )
        root = projection_node

        # 4. process additional instructions/nodes
        if len(args) > 3:
            for arg in args[3:]:
                if isinstance(arg, TimeRange):
                    filter_node.timerange = arg
                elif isinstance(arg, Limit):
                    root = self.irgraph.add_node(arg, projection_node)

        return root, output_type, native_projection_field

    def find(self, args) -> (Instruction, str, str):
        output_var_type = args[0].value
        relation = args[1].value.upper()
        if_reverse, input_var_name = (
            (True, args[3].value)
            if hasattr(args[2], "type")
            and args[2].type == self.token_prefix + "REVERSED"
            else (False, args[2].value)
        )
        input_var = self.irgraph.get_variable(input_var_name)
        input_var_type = input_var.entity_type

        if input_var_type == "event":  # event-to-entity relation
            output_projection = _get_entity_event_relation_projection(
                self.entity_event_relation_table,
                output_var_type,
                relation,
            )
            filter_node = self.irgraph.add_node(Filter(), input_var)
            root = self.irgraph.add_node(
                ProjectEntity(output_projection, output_projection), filter_node
            )
        elif output_var_type == "event":  # entity-to-event relation
            if not if_reverse:  # this relation always require `BY`
                raise UnsupportedObjectRelation(f'Missing "BY" after "{relation}"')
            input_specifier = _get_entity_event_relation_projection(
                self.entity_event_relation_table,
                input_var_type,
                relation,
            )
            filter_node = _create_filter_for_find(
                self.entity_identifier_map,
                input_var,
                input_specifier,
            )
            ds = self.irgraph.find_datasource_of_node(input_var)
            filter_node = self.irgraph.add_node(filter_node, ds)

            # still need projection for translation/mapping to take place
            root = self.irgraph.add_node(
                ProjectEntity(output_var_type, output_var_type), filter_node
            )
        else:  # entity-to-entity relation
            lookup_input_type, lookup_output_type = (
                (output_var_type, input_var_type)
                if if_reverse
                else (input_var_type, output_var_type)
            )
            input_specifier, output_projection = (
                _get_entity_entity_relation_specifier_projection(
                    self.entity_entity_relation_table,
                    lookup_input_type,
                    lookup_output_type,
                    relation,
                )
            )
            input_specifier, output_projection = (
                (output_projection, input_specifier)
                if if_reverse
                else (input_specifier, output_projection)
            )
            filter_node = _create_filter_for_find(
                self.entity_identifier_map,
                input_var,
                input_specifier,
            )
            ds = self.irgraph.find_datasource_of_node(input_var)
            filter_node = self.irgraph.add_node(filter_node, ds)
            root = self.irgraph.add_node(
                ProjectEntity(output_projection, output_projection), filter_node
            )

        if len(args) > 3:
            for arg in args[3:]:
                if isinstance(arg, Filter):
                    # merge the user-specified filter (where clause)
                    filter_node.exp = BoolExp(filter_node.exp, ExpOp.AND, arg.exp)
                if isinstance(arg, TimeRange):
                    # set user-specified time range
                    filter_node.timerange = arg
                elif isinstance(arg, Limit):
                    root = self.irgraph.add_node(arg, root)

        return root, output_var_type, output_var_type

    def apply(self, args) -> List[Return]:
        scheme, analytic_name = args[0]
        if len(args[1]) > 1:
            raise NotImplementedError("Apply on multiple variables")
        else:
            refvar = args[1][0]
        params = args[2] if len(args) > 2 else {}
        vds = AnalyticsInterface(interface=scheme)
        analytic = Analytic(name=analytic_name, params=params)
        _logger.debug("apply: analytic: %s", analytic)

        output_var = Variable(refvar.name, refvar.entity_type, refvar.native_type)

        self.irgraph.add_node(analytic, refvar)
        self.irgraph.add_node(vds)
        self.irgraph.add_edge(vds, analytic)
        self.irgraph.add_node(output_var, analytic)
        return []

    def where_clause(self, args) -> Filter:
        exp = args[0]
        return Filter(exp)

    def attr_clause(self, args) -> ProjectAttrs:
        attrs = args[0].split(",")
        attrs = tuple(attr.strip() for attr in attrs)
        return ProjectAttrs(attrs)

    def sort_clause(self, args) -> Sort:
        # args[0] is Token('BY', 'BY')
        return Sort(*args[1:])

    def expression_or(self, args) -> BoolExp:
        return BoolExp(args[0], ExpOp.OR, args[1])

    def expression_and(self, args) -> BoolExp:
        return BoolExp(args[0], ExpOp.AND, args[1])

    def comparison_std(self, args) -> FComparison:
        """Emit a Comparison object for a Filter"""
        field = args[0].value
        op = args[1]
        value = args[2]
        return _create_comp(field, op, value, True)

    def args(self, args) -> dict:
        return dict(args)

    def arg_kv_pair(self, args):
        name = args[0].value
        if isinstance(args[1], ReferenceValue):
            value = args[1].reference
        else:
            value = args[1]  # Should be int or float?
        return (name, value)

    def op(self, args):
        """Convert operator token to a plain string"""
        return " ".join([arg.upper() for arg in args])

    def op_keyword(self, args):
        """Convert operator token to a plain string"""
        return args[0].value

    # Literals
    def advanced_string(self, args) -> str:
        value = _unescape_quoted_string(args[0].value)
        return value

    def reference_or_simple_string(self, args) -> ReferenceValue:
        vname = args[0].value
        attr = args[1].value if len(args) > 1 else None
        return ReferenceValue(vname, (attr,))

    def number(self, args):
        v = args[0].value
        try:
            return int(v)
        except ValueError:
            return float(v)

    def value(self, args):
        return args[0]

    def literal_list(self, args):
        return args

    def literal(self, args):
        return args[0]

    def datasource(self, args) -> DataSource:
        return DataSource(args[0].value)

    def analytics_uri(self, args) -> (str, str):
        scheme, _, analytic = args[0].value.partition("://")
        _logger.debug("analytics_uri: %s %s", scheme, analytic)
        return scheme, analytic

    # Timespans
    def timespan_relative(self, args) -> TimeRange:
        num = int(args[0])
        unit = args[1]
        if unit == "DAY":
            delta = timedelta(days=num)
        elif unit == "HOUR":
            delta = timedelta(hours=num)
        elif unit == "MINUTE":
            delta = timedelta(minutes=num)
        elif unit == "SECOND":
            delta = timedelta(seconds=num)
        stop = datetime.now(timezone.utc)
        start = stop - delta
        return TimeRange(start, stop)

    def timespan_absolute(self, args) -> TimeRange:
        start = to_datetime(args[0])
        stop = to_datetime(args[1])
        return TimeRange(start, stop)

    def day(self, _args):
        return "DAY"

    def hour(self, _args):
        return "HOUR"

    def minute(self, _args):
        return "MINUTE"

    def second(self, _args):
        return "SECOND"

    def timestamp(self, args) -> str:
        return args[0]

    # Limit
    def limit_clause(self, args) -> Limit:
        n = int(args[0])
        return Limit(n)

    def offset_clause(self, args) -> Offset:
        n = int(args[0])
        return Offset(n)

    def disp(self, args) -> List[Return]:
        root, entity_type, native_type = args[0]
        _logger.debug("disp: root = %s", root)
        if isinstance(root, ProjectAttrs):
            root.attrs = translate_attributes_projection_to_ocsf(
                self.field_map, native_type, entity_type, root.attrs
            )
        ret = self.irgraph.add_node(Return(), root)
        return [ret]

    def explain(self, args) -> List[Return]:
        variable = self.irgraph.get_variable(args[0].value)
        explain = self.irgraph.add_node(Explain(), variable)
        ret = self.irgraph.add_node(Return(), explain)
        return [ret]

    def info(self, args) -> List[Return]:
        variable = self.irgraph.get_variable(args[0].value)
        info = self.irgraph.add_node(Information(), variable)
        ret = self.irgraph.add_node(Return(), info)
        return [ret]



import os
from uuid import uuid4

import pytest
import sqlalchemy
from pandas import DataFrame, read_csv

from kestrel.config.internal import VIRTUAL_CACHE_VAR_DATA
from kestrel.cache import SqlCache
from kestrel.cache.sql import SqlCacheVirtual
from kestrel.config import load_kestrel_config
from kestrel.frontend.parser import parse_kestrel_and_update_irgraph
from kestrel.ir.graph import IRGraph, IRGraphEvaluable
from kestrel.ir.instructions import Construct, SerializableDataFrame


@pytest.fixture
def process_creation_events():
    # return a two-node graph:
    #   - Construct: table from logs_ocsf_process_creation.csv
    #   - Variable es: events pointing to the Construct
    graph = IRGraph()
    parse_kestrel_and_update_irgraph("es = NEW event [ {'id': 1} ]", graph, {})
    data_node = graph.get_nodes_by_type(Construct)[0]
    test_dir = os.path.dirname(os.path.abspath(__file__))
    data_node.data = SerializableDataFrame(read_csv(os.path.join(test_dir, "logs_ocsf_process_creation.csv")))
    return graph


@pytest.fixture
def kestrel_config():
    return load_kestrel_config()


def test_sql_cache_set_get_del():
    c = SqlCache()
    idx = uuid4()
    df = DataFrame({'foo': [1, 2, 3]})
    c[idx] = df

    assert df.equals(c[idx])

    del c[idx]
    assert idx not in c


def test_sql_cache_constructor():
    ids = [uuid4() for i in range(5)]
    df = DataFrame({'foo': [1, 2, 3]})
    c = SqlCache({x:df for x in ids})
    for u in ids:
        assert df.equals(c[u])
    for u in ids:
        del c[u]
        assert u not in c


def test_eval_new_disp():
    stmt = """
proclist = NEW process [ {"name": "cmd.exe", "pid": 123}
                       , {"name": "explorer.exe", "pid": 99}
                       , {"name": "firefox.exe", "pid": 201}
                       , {"name": "chrome.exe", "pid": 205}
                       ]
DISP proclist ATTR name
"""
    graph = IRGraph()
    rets = parse_kestrel_and_update_irgraph(stmt, graph, {})
    graph = IRGraphEvaluable(graph)
    c = SqlCache()
    mapping = c.evaluate_graph(graph, c)

    # check the return is correct
    assert len(rets) == 1
    df = mapping[rets[0].id]
    assert df.to_dict("records") == [ {"name": "cmd.exe"}
                                    , {"name": "explorer.exe"}
                                    , {"name": "firefox.exe"}
                                    , {"name": "chrome.exe"}
                                    ]


def test_eval_new_filter_disp():
    stmt = """
proclist = NEW process [ {"name": "cmd.exe", "pid": 123}
                       , {"name": "explorer.exe", "pid": 99}
                       , {"name": "firefox.exe", "pid": 201}
                       , {"name": "chrome.exe", "pid": 205}
                       ]
browsers = proclist WHERE name = 'firefox.exe' OR name = 'chrome.exe'
DISP browsers ATTR name, pid
"""
    graph = IRGraph()
    rets = parse_kestrel_and_update_irgraph(stmt, graph, {})
    graph = IRGraphEvaluable(graph)
    c = SqlCache()
    mapping = c.evaluate_graph(graph, c)

    # check the return is correct
    assert len(rets) == 1
    df = mapping[rets[0].id]
    assert df.to_dict("records") == [ {"name": "firefox.exe", "pid": 201}
                                    , {"name": "chrome.exe", "pid": 205}
                                    ]

    
def test_eval_two_returns():
    stmt = """
proclist = NEW process [ {"name": "cmd.exe", "pid": 123}
                       , {"name": "explorer.exe", "pid": 99}
                       , {"name": "firefox.exe", "pid": 201}
                       , {"name": "chrome.exe", "pid": 205}
                       ]
browsers = proclist WHERE name != "cmd.exe"
DISP browsers
DISP browsers ATTR pid
"""
    graph = IRGraph()
    rets = parse_kestrel_and_update_irgraph(stmt, graph, {})
    graph = IRGraphEvaluable(graph)
    c = SqlCache()

    # first DISP
    gs = graph.find_dependent_subgraphs_of_node(rets[0], c)
    assert len(gs) == 1
    mapping = c.evaluate_graph(gs[0], c)
    df1 = DataFrame([ {"name": "explorer.exe", "pid": 99}
                    , {"name": "firefox.exe", "pid": 201}
                    , {"name": "chrome.exe", "pid": 205}
                    ])
    assert len(mapping) == 1
    assert df1.equals(mapping[rets[0].id])

    # second DISP
    gs = graph.find_dependent_subgraphs_of_node(rets[1], c)
    assert len(gs) == 1
    mapping = c.evaluate_graph(gs[0], c)
    df2 = DataFrame([ {"pid": 99}
                    , {"pid": 201}
                    , {"pid": 205}
                    ])
    assert len(mapping) == 1
    assert df2.equals(mapping[rets[1].id])


def test_issue_446():
    """The `WHERE name IN ...` below was raising `sqlalchemy.exc.StatementError: (builtins.KeyError) 'name_1'`
    https://github.com/opencybersecurityalliance/kestrel-lang/issues/446
    """
    stmt = """
proclist = NEW process [ {"name": "cmd.exe", "pid": 123}
                       , {"name": "explorer.exe", "pid": 99}
                       , {"name": "firefox.exe", "pid": 201}
                       , {"name": "chrome.exe", "pid": 205}
                       ]
browsers = proclist WHERE name IN ("explorer.exe", "firefox.exe", "chrome.exe")
"""
    graph = IRGraph()
    rets = parse_kestrel_and_update_irgraph(stmt, graph, {})
    graph = IRGraphEvaluable(graph)
    c = SqlCache()
    _ = c.evaluate_graph(graph, c)


def test_eval_filter_with_ref():
    stmt = """
proclist = NEW process [ {"name": "cmd.exe", "pid": 123}
                       , {"name": "explorer.exe", "pid": 99}
                       , {"name": "firefox.exe", "pid": 201}
                       , {"name": "chrome.exe", "pid": 205}
                       ]
browsers = proclist WHERE name = 'firefox.exe' OR name = 'chrome.exe'
specials = proclist WHERE pid IN [123, 201]
p2 = proclist WHERE pid = browsers.pid and name = specials.name
DISP p2 ATTR name, pid
"""
    graph = IRGraph()
    rets = parse_kestrel_and_update_irgraph(stmt, graph, {})
    graph = IRGraphEvaluable(graph)
    c = SqlCache()
    mapping = c.evaluate_graph(graph, c)

    # check the return is correct
    assert len(rets) == 1
    df = mapping[rets[0].id]
    assert df.to_dict("records") == [ {"name": "firefox.exe", "pid": 201} ]


def test_get_virtual_copy():
    stmt = """
proclist = NEW process [ {"name": "cmd.exe", "pid": 123}
                       , {"name": "explorer.exe", "pid": 99}
                       , {"name": "firefox.exe", "pid": 201}
                       , {"name": "chrome.exe", "pid": 205}
                       ]
browsers = proclist WHERE name = 'firefox.exe' OR name = 'chrome.exe'
"""
    graph = IRGraph()
    rets = parse_kestrel_and_update_irgraph(stmt, graph, {})
    graph = IRGraphEvaluable(graph)
    c = SqlCache()
    mapping = c.evaluate_graph(graph, c)
    v = c.get_virtual_copy()
    new_entry = uuid4()
    v[new_entry] = DataFrame()

    # v[new_entry] calls the right method
    assert isinstance(v, SqlCacheVirtual)
    assert v[new_entry].equals(VIRTUAL_CACHE_VAR_DATA)

    # the two cache_catalog are different
    assert new_entry not in c
    assert new_entry in v
    del v[new_entry]
    assert new_entry not in v
    for u in c:
        del v[u]
    assert len(v) == 0
    assert len(c) == 1


def test_eval_find_event_to_entity(process_creation_events):
    graph = process_creation_events
    stmt = "procs = FIND process RESPONDED es WHERE device.os = 'Linux' DISP procs"
    rets = parse_kestrel_and_update_irgraph(stmt, graph, {})
    graph = IRGraphEvaluable(graph)
    c = SqlCache()
    mapping = c.evaluate_graph(graph, c)
    assert len(rets) == 1
    df = mapping[rets[0].id]
    assert list(df.columns) == ['cmd_line', 'name', 'pid', 'uid', 'endpoint.uid', 'endpoint.name',
       'endpoint.os', 'file.name', 'file.path', 'user.uid', 'user.name',
       'user.type_id', 'parent_process.cmd_line', 'parent_process.name',
       'parent_process.pid', 'parent_process.uid',
       'parent_process.endpoint.uid', 'parent_process.endpoint.name',
       'parent_process.endpoint.os', 'user.endpoint.uid', 'user.endpoint.name',
       'user.endpoint.os', 'file.endpoint.uid', 'file.endpoint.name',
       'file.endpoint.os']
    assert df.shape[0] == 5  # WHERE clause filtered out 4 out of 9, so 5 remains


def test_eval_find_entity_to_event(process_creation_events, kestrel_config):
    graph = process_creation_events
    stmt = """
        procs = FIND process RESPONDED es WHERE device.os = 'Linux'
        eves = FIND event ORIGINATED BY procs
        DISP eves
    """
    rets = parse_kestrel_and_update_irgraph(stmt, graph, kestrel_config["entity_identifier"])
    graph = IRGraphEvaluable(graph)
    c = SqlCache()
    mapping = c.evaluate_graph(graph, c)
    assert len(rets) == 1
    df = mapping[rets[0].id]

    # 1. WHERE clause filtered out 4 out of 9, so 5 remains
    # 2. In the 5, 4 are not parent process of others, only the first is parent process
    # 3. There are 4 lines of logs/events that uses the parent process
    assert df.shape[0] == 4
    assert df.shape[1] == 48  # full event: the number of columns in the csv


def test_eval_find_entity_to_entity(process_creation_events, kestrel_config):
    graph = process_creation_events
    stmt = """
        procs = FIND process RESPONDED es WHERE device.os = 'Linux'
        parents = FIND process CREATED procs
        DISP parents
    """
    rets = parse_kestrel_and_update_irgraph(stmt, graph, kestrel_config["entity_identifier"])
    graph = IRGraphEvaluable(graph)
    c = SqlCache()
    mapping = c.evaluate_graph(graph, c)
    assert len(rets) == 1
    df = mapping[rets[0].id]

    # 1. WHERE clause filtered out 4 out of 9, so 5 remains
    # 2. The last 4 share the same parent
    # 3. So there are 2 processes returned/displayed after dedup
    assert df.shape[0] == 2
    assert list(df.columns) == ['cmd_line', 'name', 'pid', 'uid', 'endpoint.uid', 'endpoint.name', 'endpoint.os']


def test_explain_find_event_to_entity(process_creation_events):
    graph = process_creation_events
    stmt = "procs = FIND process RESPONDED es WHERE device.os = 'Linux' EXPLAIN procs"
    rets = parse_kestrel_and_update_irgraph(stmt, graph, {})
    graph = IRGraphEvaluable(graph)
    c = SqlCache()
    mapping = c.explain_graph(graph, c)
    assert len(rets) == 1
    explanation = mapping[rets[0].id]
    construct = graph.get_nodes_by_type(Construct)[0]
    stmt = explanation.action.statement.replace('"', '')
    assert stmt == f"""WITH es AS 
(SELECT DISTINCT * 
FROM {construct.id.hex}), 
procs AS 
(SELECT DISTINCT process.cmd_line AS cmd_line, process.name AS name, process.pid AS pid, process.uid AS uid, process.endpoint.uid AS endpoint.uid, process.endpoint.name AS endpoint.name, process.endpoint.os AS endpoint.os, process.file.name AS file.name, process.file.path AS file.path, process.user.uid AS user.uid, process.user.name AS user.name, process.user.type_id AS user.type_id, process.parent_process.cmd_line AS parent_process.cmd_line, process.parent_process.name AS parent_process.name, process.parent_process.pid AS parent_process.pid, process.parent_process.uid AS parent_process.uid, process.parent_process.endpoint.uid AS parent_process.endpoint.uid, process.parent_process.endpoint.name AS parent_process.endpoint.name, process.parent_process.endpoint.os AS parent_process.endpoint.os, process.user.endpoint.uid AS user.endpoint.uid, process.user.endpoint.name AS user.endpoint.name, process.user.endpoint.os AS user.endpoint.os, process.file.endpoint.uid AS file.endpoint.uid, process.file.endpoint.name AS file.endpoint.name, process.file.endpoint.os AS file.endpoint.os 
FROM es 
WHERE device.os = 'Linux')
 SELECT DISTINCT * 
FROM procs"""


def test_eval_information():
    stmt = """
events = NEW event [ {"process.name": "cmd.exe", "process.pid": 123, "user.name": "user", "event_type": "process"} ]
INFO events
"""
    graph = IRGraph()
    rets = parse_kestrel_and_update_irgraph(stmt, graph, {})
    graph = IRGraphEvaluable(graph)
    c = SqlCache()
    mapping = c.evaluate_graph(graph, c)

    # check the return is correct
    assert len(rets) == 1
    df = mapping[rets[0].id]
    attrs = df["attributes"].to_list()
    assert attrs == ['event_type', 'process.name, process.pid', 'user.name']



import json
from collections import Counter

import pytest

from kestrel.frontend.parser import parse_kestrel_and_update_irgraph
from kestrel.ir.filter import (AbsoluteTrue, BoolExp, ExpOp, FloatComparison,
                               IntComparison, ListComparison, ListOp,
                               MultiComp, NumCompOp, RefComparison,
                               ReferenceValue, StrComparison, StrCompOp,
                               get_references_from_exp,
                               resolve_reference_with_function)
from kestrel.ir.graph import IRGraph
from kestrel.ir.instructions import (Construct, Filter, ProjectAttrs, Variable,
                                     instruction_from_json)


@pytest.mark.parametrize(
    "field, op, value", [
        ("foo", StrCompOp.EQ, "bar"),
        ("foo", NumCompOp.EQ, 42),
        ("foo", NumCompOp.EQ, 3.14),
        ("foo", StrCompOp.NEQ, "bar"),
        ("foo", NumCompOp.NEQ, 42),
        ("foo", NumCompOp.NEQ, 3.14),
        ("foo", StrCompOp.LIKE, "%bar"),
        ("foo", StrCompOp.NLIKE, "%bar"),
    ]
)
def test_comparison(field, op, value):
    if isinstance(value, int):
        comp = IntComparison(field=field, op=op, value=value)
    elif isinstance(value, float):
        comp = FloatComparison(field=field, op=op, value=value)
    else:
        comp = StrComparison(field=field, op=op, value=value)
    assert comp.field == field
    assert comp.op == op
    assert comp.value == value
    json_data: str = comp.to_json()
    data: dict = json.loads(json_data)
    assert data["field"] == field
    assert data["op"] == op
    assert data["value"] == value
    if isinstance(value, int):
        comp2 = IntComparison.from_json(json_data)
    elif isinstance(value, float):
        comp2 = FloatComparison.from_json(json_data)
    else:
        comp2 = StrComparison.from_json(json_data)
    assert comp == comp2


@pytest.mark.parametrize(
    "field, op, value", [
        ("foo", ListOp.IN, ["a", "b", "c"]),
        ("foo", ListOp.NIN, ["a", "b", "c"]),
        ("foo", ListOp.IN, [1, 2, 3]),
        ("foo", ListOp.NIN, [1, 2, 3]),
    ]
)
def test_list_comparison(field, op, value):
    comp = ListComparison(field=field, op=op, value=value)
    assert comp.field == field
    assert comp.op == op
    assert comp.value == value
    json_data: str = comp.to_json()
    data: dict = json.loads(json_data)
    assert data["field"] == field
    assert data["op"] == op
    assert data["value"] == value
    comp2 = ListComparison.from_json(json_data)
    assert comp == comp2




def test_multi_comparison():
    comp1 = StrComparison("foo", StrCompOp.EQ, "X")
    comp2 = StrComparison("bar", StrCompOp.EQ, "Y")
    comp3 = StrComparison("baz", StrCompOp.EQ, "Z")
    mcomp = MultiComp(ExpOp.OR, [comp1, comp2, comp3])
    data = mcomp.to_json()
    mcomp2 = MultiComp.from_json(data)
    assert mcomp == mcomp2


@pytest.mark.parametrize(
    "lhs, op, rhs", [
        (StrComparison("foo", StrCompOp.EQ, "bar"), ExpOp.AND, AbsoluteTrue()),
        (StrComparison("foo", StrCompOp.EQ, "bar"), ExpOp.AND, IntComparison("baz", NumCompOp.EQ, 42)),
        (StrComparison("foo", StrCompOp.LIKE, "%bar%"), ExpOp.OR, IntComparison("baz", NumCompOp.LE, 42)),
        (IntComparison("baz", NumCompOp.GE, 42), ExpOp.AND, StrComparison("foo", StrCompOp.NEQ, "bar")),
        (IntComparison("baz", NumCompOp.NEQ, 42), ExpOp.OR, StrComparison("foo", StrCompOp.EQ, "bar")),
        (StrComparison("foo", StrCompOp.EQ, "bar"), ExpOp.AND, ListComparison("baz", ListOp.IN, ["a", "b", "c"])),
        (StrComparison("foo", StrCompOp.EQ, "bar"), ExpOp.OR, ListComparison("baz", ListOp.IN, [1, 2, 3])),
        (ListComparison("baz", ListOp.IN, ["a", "b", "c"]), ExpOp.AND, StrComparison("foo", StrCompOp.EQ, "bar")),
        (ListComparison("baz", ListOp.IN, [1, 2, 3]), ExpOp.OR, StrComparison("foo", StrCompOp.EQ, "bar")),
        (StrComparison("foo", StrCompOp.EQ, "X"), ExpOp.AND,
         MultiComp(ExpOp.OR, [StrComparison("bar", StrCompOp.EQ, "A"), StrComparison("baz", StrCompOp.EQ, "B")])),
    ]
)
def test_bool_exp(lhs, op, rhs):
    exp = BoolExp(lhs, op, rhs)
    data = exp.to_json()
    exp2 = BoolExp.from_json(data)
    assert exp == exp2

    # Also test Filter
    filt = Filter(exp)
    data = filt.to_json()
    filt2 = instruction_from_json(data)
    assert filt == filt2


def test_filter_compound_exp():
    comp1 = StrComparison("foo", StrCompOp.EQ, "bar")
    comp2 = IntComparison("baz", NumCompOp.EQ, 42)
    exp1 = BoolExp(comp1, ExpOp.AND, comp2)
    comp3 = StrComparison("thing1", StrCompOp.NEQ, "abc")
    comp4 = ListComparison("thing2", ListOp.IN, [1, 2, 3])
    exp2 = BoolExp(comp3, ExpOp.OR, comp4)
    exp3 = BoolExp(exp1, ExpOp.AND, exp2)
    filt = Filter(exp3)
    data = filt.to_json()
    filt2 = instruction_from_json(data)
    assert filt == filt2


def test_filter_with_reference():
    stmt = """
x = NEW process [ {"name": "cmd.exe", "pid": 123} ]
y = NEW process [ {"asdf": "abc.exe"} ]
z = x WHERE foo = 'bar' OR baz = y.asdf
"""
    graph = IRGraph()
    parse_kestrel_and_update_irgraph(stmt, graph, {})
    assert len(graph.nodes()) == 7
    assert Counter(map(type, graph.nodes())) == Counter([Variable, Variable, Variable, ProjectAttrs, Filter, Construct, Construct])
    filter_nodes = graph.get_nodes_by_type(Filter)
    exp = filter_nodes[0].exp
    exp_dict = exp.to_dict()
    assert exp_dict == {'lhs': {'field': 'foo', 'op': '=', 'value': 'bar'}, 'op': 'OR', 'rhs': {'fields': ['baz'], 'op': 'IN', 'value': {'reference': 'y', 'attributes': ['asdf']}}}


def test_fill_references_in_exp():
    lhs = StrComparison("foo", StrCompOp.EQ, "bar")
    rhs = RefComparison(["baz"], "=", ReferenceValue("var", ("attr",)))
    exp = BoolExp(lhs, ExpOp.AND, rhs)
    rs = get_references_from_exp(exp)
    assert len(list(rs)) == 1
    resolve_reference_with_function(exp, lambda x: 5)
    assert exp.rhs.value == 5



import json

import pandas as pd
import pytest

from kestrel.config.utils import load_kestrel_config
from kestrel.exceptions import IncompleteDataMapping
from kestrel.mapping.data_model import (
    check_entity_identifier_existence_in_mapping, load_default_mapping,
    reverse_mapping, translate_comparison_to_native,
    translate_comparison_to_ocsf, translate_dataframe,
    translate_projection_to_native)

# A "custom" mapping for an opensearch/elasticsearch datasource.
# This mapping works with data from Blue Team Village's 2023 DefCon CTF, for example.
WINLOGBEAT_MAPPING = {
    "type_uid": {
        "native_field": "winlog.event_id",
        "native_value": {
            100701: ["1", "4688"],  # process launch
            100702: ["5", "4689"],  # process terminate
            300201: ["4624", "4625"], # auth: logon (success and failure)
        },
        "XXX_ocsf_value": {
            "1": 100701,  # Sysmon process create
            "5": 100702,  # Sysmon process terminate
            "4624": 300201,  # Security logon success
            "4625": 300201,  # Security logon failure
            "4688": 100701,  # Security process create
            "4689": 100702,  # Security process terminate
        }
    },
    "file": {
        "path": "file.path",
        "name": "file.name"
    },
    "process": {
        "cmd_line": "winlog.event_data.CommandLine",
        "pid": {
            "native_field": "winlog.event_data.ProcessId",
            "native_value": "to_str",
            "ocsf_value": "to_int"
        },
        "uid": "winlog.event_data.ProcessGuid",
        "file": {
            "path": "winlog.event_data.Image",
            "name": [
                {
                    "native_field": "winlog.event_data.Image",
                    "native_op": "LIKE",
                    "native_value": "winpath_endswith",
                    "ocsf_value": "basename"
                },
                {   # Only for testing - winlogbeat will only use windows paths
                    "native_field": "winlog.event_data.Image",
                    "native_op": "LIKE",
                    "native_value": "posixpath_endswith",
                    "ocsf_value": "basename"
                }
            ],
            "parent_folder": [
                {
                    "native_field": "winlog.event_data.Image",
                    "native_op": "LIKE",
                    "native_value": "winpath_startswith",
                    "ocsf_value": "dirname"
                }
            ]
        },
        "parent_process": {
            "cmd_line": "winlog.event_data.ParentCommandLine",
            "pid": "winlog.event_data.ParentProcessId",
            "uid": "winlog.event_data.ParentProcessGuid",
            "file": {
                "path": "winlog.event_data.ParentImage",
                "name": [
                    {
                        "native_field": "winlog.event_data.ParentImage",
                        "native_op": "LIKE",
                        "native_value": "winpath_endswith",
                        "ocsf_value": "basename"
                    }
                ],
                "parent_folder": [
                    {
                        "native_field": "winlog.event_data.ParentImage",
                        "native_op": "LIKE",
                        "native_value": "winpath_startswith",
                        "ocsf_value": "dirname"
                    }
                ]
            }
        }
    },
    "dst_endpoint": {
        "ip": "winlog.event_data.DestinationIp",
        "port": "winlog.event_data.DestinationPort"
    },
    "src_endpoint": {
        "ip": "winlog.event_data.SourceIp",
        "port": "winlog.event_data.SourcePort"
    }
}


# Mapping for testing missing identifier
INCOMPLETE_MAPPING = {
    "process": {
        "pid": "process.pid"
    }
}


# Simplified subset of the standard mapping
STIX_MAPPING = {
    "device": {
        "ip": "ipv4-addr:value"
    },
    "endpoint": {
        "ip": "ipv4-addr:value"
    },
}


# This mapping is used in 2 places:
# - frontend comparison from ECS to OCSF
# - backend comparison from OCSF to ECS (datasource)
ECS_MAPPING = load_default_mapping("ecs")


def test_reverse_mapping_ipv4():
    reverse_map = reverse_mapping(STIX_MAPPING)
    ipv4 = reverse_map["ipv4-addr:value"]
    assert isinstance(ipv4, list)
    assert set(ipv4) == {"device.ip", "endpoint.ip"}


def test_reverse_mapping_executable():
    reverse_map = reverse_mapping(ECS_MAPPING)
    exe = reverse_map["process.executable"]
    assert isinstance(exe, list)
    assert "process.file.path" in exe
    for item in exe:
        if isinstance(item, dict):
            assert "ocsf_field" in item
            if item["ocsf_field"] == "process.file.name":
                # Make sure all metadata from the mapping got reversed
                assert item["native_value"] in ("posixpath_endswith", "winpath_endswith")
                assert item["native_op"] == "LIKE"
                assert item["ocsf_value"] == "basename"


def test_reverse_mapping_event_id():
    rmap = reverse_mapping(WINLOGBEAT_MAPPING)
    assert rmap["winlog.event_id"][0]["ocsf_value"]["1"] == [100701]
    assert rmap["winlog.event_id"][0]["ocsf_value"]["5"] == [100702]
    assert rmap["winlog.event_id"][0]["ocsf_value"]["4688"] == [100701]
    assert rmap["winlog.event_id"][0]["ocsf_value"]["4689"] == [100702]


@pytest.mark.parametrize(
    "dmm, field, op, value, expected_result",
    [
        (WINLOGBEAT_MAPPING, "process.file.path", "=", "C:\\TMP\\foo.exe",
         [("winlog.event_data.Image", "=", "C:\\TMP\\foo.exe")]),
        (WINLOGBEAT_MAPPING, "process.file.name", "=", "foo.exe",
         [("winlog.event_data.Image", "LIKE", "%\\foo.exe"),
          ("winlog.event_data.Image", "LIKE", "%/foo.exe")]),
        (ECS_MAPPING, "process.file.path", "=", "C:\\TMP\\foo.exe",
         [("process.executable", "=", "C:\\TMP\\foo.exe")]),
        (ECS_MAPPING, "process.file.name", "=", "foo.exe",
         [("process.executable", "LIKE", "%\\foo.exe"),
          ("process.executable", "LIKE", "%/foo.exe")]),
        (WINLOGBEAT_MAPPING, "type_uid", "=", 100701,
         [("winlog.event_id", "IN", ["1", "4688"])]),
        (ECS_MAPPING, "type_uid", "=", 100701,
         [("event.code", "IN", ["1", "4688"])]),
        (ECS_MAPPING, "type_uid", "=", 300201,
         [("event.code", "IN", ["4624", "4625"])]),
        (ECS_MAPPING, "type_uid", "!=", 300201,
         [("event.code", "NOT IN", ["4624", "4625"])]),
    ],
)
def test_translate_comparison_to_native(dmm, field, op, value, expected_result):
    assert translate_comparison_to_native(dmm, field, op, value) == expected_result


@pytest.mark.parametrize(
    "dmm, field, op, value, expected_result",
    [
        (ECS_MAPPING, "process.executable", "=", "C:\\TMP\\foo.exe",
         [
            ("process.file.path", "=", "C:\\TMP\\foo.exe"),
            ("process.file.name", "=", "foo.exe"),
            ("process.file.parent_folder", "=", "C:\\TMP"),
         ]),
        (ECS_MAPPING, "process.executable", "LIKE", "%\\foo.exe",
         [
            ("process.file.path", "LIKE", "%\\foo.exe"),
            ("process.file.name", "LIKE", "foo.exe"),     #TODO: could optimize this to "="
            ("process.file.parent_folder", "LIKE", "%"),  #TODO: could eliminate this?
         ]),
        (STIX_MAPPING, "ipv4-addr:value", "=", "198.51.100.13",
         [
             ("device.ip", "=", "198.51.100.13"),
             ("endpoint.ip", "=", "198.51.100.13"),
         ]),
    ],
)
def test_translate_comparison_to_ocsf(dmm, field, op, value, expected_result):
    """Test the translate function."""
    reverse_dmm = reverse_mapping(dmm)  # Make the dmms fixtures?
    assert set(translate_comparison_to_ocsf(reverse_dmm, field, op, value)) == set(expected_result)


@pytest.mark.parametrize(
    "dmm, entity, field, expected_result",
    [
        (WINLOGBEAT_MAPPING, "process", ["file.name", "pid"],
         [("winlog.event_data.Image", "file.name"), ("winlog.event_data.ProcessId", "pid")]),
        (WINLOGBEAT_MAPPING, "process", None,
         [("winlog.event_data.CommandLine", "cmd_line"),
          ("winlog.event_data.ProcessId", "pid"),
          ("winlog.event_data.ProcessGuid", "uid"),
          ("winlog.event_data.Image", "file.path"),
          ("winlog.event_data.Image", "file.name"),
          ("winlog.event_data.Image", "file.parent_folder"),
          ("winlog.event_data.ParentCommandLine", "parent_process.cmd_line"),
          ("winlog.event_data.ParentProcessId", "parent_process.pid"),
          ("winlog.event_data.ParentProcessGuid", "parent_process.uid"),
          ("winlog.event_data.ParentImage", "parent_process.file.path"),
          ("winlog.event_data.ParentImage", "parent_process.file.name"),
          ("winlog.event_data.ParentImage", "parent_process.file.parent_folder"),
         ]),
    ],
)
def test_translate_projection_to_native(dmm, entity, field, expected_result):
    assert translate_projection_to_native(dmm, entity, field) == expected_result


def test_translate_dataframe():  #TODO: more testing here
    df = pd.DataFrame({"file.path": [r"C:\Windows\System32\cmd.exe", r"C:\TMP"],
                       "pid": [1, 2]})
    dmm = load_default_mapping("ecs")
    df = translate_dataframe(df, dmm["process"])
    #TODO:assert df["file.name"].iloc[0] == "cmd.exe"


def test_translate_dataframe_events():
    df = pd.DataFrame(
        {
            "process.file.path": [r"C:\Windows\System32\cmd.exe", r"C:\TMP"],
            "process.pid": [1, 2],
            "type_uid": ["4688", "1234"],
        }
    )
    df = translate_dataframe(df, WINLOGBEAT_MAPPING)
    assert df["type_uid"].iloc[0] == 100701
    assert df["type_uid"].iloc[1] == "1234"  # Passthrough?


def test_incomplete_mapping_no_identifier():
    identifier_config = load_kestrel_config()["entity_identifier"]
    with pytest.raises(IncompleteDataMapping):
        check_entity_identifier_existence_in_mapping(INCOMPLETE_MAPPING, identifier_config, "test interface")



import json
from abc import ABC, abstractmethod
from typing import Any, Iterable, List, Mapping, MutableMapping, Optional
from uuid import UUID

from kestrel.display import GraphletExplanation
from kestrel.exceptions import InvalidSerializedDatasourceInterfaceCacheCatalog
from kestrel.ir.graph import IRGraphEvaluable
from kestrel.ir.instructions import Instruction
from pandas import DataFrame

MODULE_PREFIX = "kestrel_interface_"


class AbstractInterface(ABC):
    """Abstract class for datasource/analytics interface

    Concepts:

    - Think an interface as a type of datalakes

    - Think a storage as a datalake

    - Think a datasource as a table in the datalake

    Attributes:

        session_id: the optional information to derive table name in datalake

        cache_catalog: map a cached item (instruction.id) to datalake table/view name
    """

    def __init__(
        self,
        serialized_cache_catalog: Optional[str] = None,
        session_id: Optional[UUID] = None,
    ):
        self.session_id = session_id
        self.cache_catalog: MutableMapping[UUID, str] = {}

        if serialized_cache_catalog:
            try:
                self.cache_catalog = json.loads(serialized_cache_catalog)
            except:
                raise InvalidSerializedDatasourceInterfaceCacheCatalog()

    # Python 3.13 will drop chain of @classmethod and @property
    # use @staticmethod instead (cannot make it a property)
    @staticmethod
    @abstractmethod
    def schemes() -> Iterable[str]:
        """The schemes to specify the interface

        Each scheme should be defined as ``("_"|LETTER) ("_"|LETTER|DIGIT)*``
        """
        ...

    @abstractmethod
    def get_datasources(self) -> List[str]:
        """Get the list of datasource names registered at this interface"""
        ...

    @abstractmethod
    def get_storage_of_datasource(self, datasource: str) -> str:
        """Get the storage name of a given datasource"""
        ...

    @abstractmethod
    def store(
        self,
        instruction_id: UUID,
        data: DataFrame,
    ):
        """Create a new table in the datalake from a dataframe

        The name of the table is a function of instruction_id (and session_id)
        in case there are conflicting tables in the datalake.

        The function can be implemented as a hashtable. If the hash collides
        with an existing hash, figure out whether the existing hash/table is
        used by the current interface and session. If yes, then replace; if
        not, then generate a new random value and record in self.cache_catalog.

        This method will update self.cache_catalog.

        Parameters:

            instruction_id: the key to be placed in `self.cache_catalog`

            data: the dataframe to store
        """
        ...

    @abstractmethod
    def evaluate_graph(
        self,
        graph: IRGraphEvaluable,
        cache: MutableMapping[UUID, Any],
        instructions_to_evaluate: Optional[Iterable[Instruction]] = None,
    ) -> Mapping[UUID, DataFrame]:
        """Evaluate the IRGraph

        Parameters:

            graph: The evaluate IRGraph

            instructions_to_evaluate: instructions to evaluate and return; by default, it will be all Return instructions in the graph

        Returns:

            DataFrames for each instruction in instructions_to_evaluate.
        """
        ...

    @abstractmethod
    def explain_graph(
        self,
        graph: IRGraphEvaluable,
        instructions_to_explain: Optional[Iterable[Instruction]] = None,
    ) -> Mapping[UUID, GraphletExplanation]:
        """Explain how to evaluate the IRGraph

        Parameters:

            graph: The evaluable IRGraph

            instructions_to_explain: instructions to explain and return; by default, it will be all Return instructions in the graph

        Returns:

            GraphletExplanation (a Kestrel Display object) for each instruction in instructions_to_explain.
        """
        ...

    def cache_catalog_to_json(self) -> str:
        """Serialize the cache catalog to a JSON string"""
        return json.dumps(self.cache_catalog)


class DatasourceInterface(AbstractInterface): ...


class AnalyticsInterface(AbstractInterface): ...



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Any, Callable, Iterable, List, Optional, Tuple, Union

from kestrel.exceptions import MismatchedFieldValueInMultiColumnComparison
from mashumaro.mixins.json import DataClassJSONMixin
from typeguard import typechecked


class NumCompOp(str, Enum):
    """Numerical comparison operators (for int and float)"""

    EQ = "="
    NEQ = "!="
    LT = "<"
    LE = "<="
    GT = ">"
    GE = ">="


@dataclass
class IntComparison(DataClassJSONMixin):
    """Integer comparison expression"""

    field: str
    op: NumCompOp
    value: int


@dataclass
class FloatComparison(DataClassJSONMixin):
    """Floating point comparison expression"""

    field: str
    op: NumCompOp
    value: float


class StrCompOp(str, Enum):
    """String comparison operators"""

    EQ = "="
    NEQ = "!="
    LIKE = "LIKE"
    NLIKE = "NOT LIKE"
    MATCHES = "MATCHES"
    NMATCHES = "NOT MATCHES"


@dataclass
class StrComparison(DataClassJSONMixin):
    """String comparison expression"""

    field: str
    op: StrCompOp
    value: str


class ListOp(str, Enum):
    """List membership operator"""

    IN = "IN"
    NIN = "NOT IN"


@dataclass
class ListStrComparison(DataClassJSONMixin):
    """List of strings membership comparison expression"""

    field: str
    op: ListOp
    value: List[str]


@dataclass
class ListIntComparison(DataClassJSONMixin):
    """List of ints membership comparison expression"""

    field: str
    op: ListOp
    value: List[int]


@dataclass
class ListComparison(DataClassJSONMixin):
    """List membership comparison expression"""

    field: str
    op: ListOp
    value: Union[List[int], List[str]]


# frozen=True for generating __hash__() method
@dataclass(frozen=True)
class ReferenceValue(DataClassJSONMixin):
    """Value for reference"""

    reference: str
    attributes: Tuple[str]


@dataclass
class RefComparison(DataClassJSONMixin):
    """Referred variable comparison"""

    fields: List[str]
    op: ListOp
    value: ReferenceValue

    def __post_init__(self):
        if len(self.fields) != len(self.value.attributes):
            raise MismatchedFieldValueInMultiColumnComparison(self)


class ExpOp(str, Enum):
    """Boolean expression operator"""

    AND = "AND"
    OR = "OR"


@dataclass
class MultiComp(DataClassJSONMixin):
    """Boolean expression of multiple comparisons.

    The single operator applies to ALL comparisons, so `OR` acts like `any` and `AND` acts like `all`.
    """

    op: ExpOp
    comps: List[
        Union[
            IntComparison, FloatComparison, StrComparison, ListComparison, RefComparison
        ]
    ]


@dataclass
class BoolExp(DataClassJSONMixin):
    """Binary boolean expression of comparisons"""

    lhs: FExpression
    op: ExpOp
    rhs: FExpression


@dataclass(frozen=True)
class AbsoluteTrue(DataClassJSONMixin):
    pass


@dataclass
class TimeRange(DataClassJSONMixin):
    """The time range of interest"""

    start: Optional[datetime] = None
    stop: Optional[datetime] = None


FExpression = Union[
    IntComparison,
    FloatComparison,
    StrComparison,
    ListComparison,
    RefComparison,
    MultiComp,
    BoolExp,
    AbsoluteTrue,  # need to place at last to avoid deserliazation issue
]


FComparison = Union[
    IntComparison,
    FloatComparison,
    StrComparison,
    ListComparison,
    RefComparison,
    MultiComp,
]


FBasicComparison = Union[
    IntComparison,
    FloatComparison,
    StrComparison,
    ListComparison,
    RefComparison,
]


@typechecked
def get_references_from_exp(exp: FExpression) -> Iterable[ReferenceValue]:
    if isinstance(exp, RefComparison):
        if isinstance(exp.value, ReferenceValue):
            # if not already resolved
            yield exp.value
    elif isinstance(exp, BoolExp):
        yield from get_references_from_exp(exp.lhs)
        yield from get_references_from_exp(exp.rhs)
    elif isinstance(exp, MultiComp):
        for comp in exp.comps:
            yield from get_references_from_exp(comp)


@typechecked
def resolve_reference_with_function(
    exp: FExpression, f: Callable[[ReferenceValue], Any]
):
    if isinstance(exp, RefComparison):
        if isinstance(exp.value, ReferenceValue):
            # if not already resolved
            exp.value = f(exp.value)
    elif isinstance(exp, BoolExp):
        resolve_reference_with_function(exp.lhs, f)
        resolve_reference_with_function(exp.rhs, f)
    elif isinstance(exp, MultiComp):
        for comp in exp.comps:
            resolve_reference_with_function(comp, f)



import pandas as pd
import pytest

from kestrel.mapping.transformers import (run_transformer,
                                          run_transformer_on_series)


@pytest.mark.parametrize(
    "transform, value, expected", [
        ("dirname", "/tmp", "/"),
        ("basename", "/tmp", "tmp"),
        ("dirname", "/usr/local/bin/thing", "/usr/local/bin"),
        ("basename", "/usr/local/bin/thing", "thing"),
        ("posixpath_startswith", "/var", "/var/%"),
        ("posixpath_endswith", "tmp.sh", r"%/tmp.sh"),
        ("dirname", r"C:\Windows\System32\cmd.exe", r"C:\Windows\System32"),
        ("basename", r"C:\Windows\System32\cmd.exe", r"cmd.exe"),
        ("winpath_startswith", r"C:\Windows\System32", r"C:\Windows\System32\%"),
        ("winpath_endswith", "cmd.exe", r"%\cmd.exe"),
        ("to_int", 1234, 1234),
        ("to_int", 1234.1234, 1234),  # Maybe this should fail?
        ("to_int", "1234", 1234),
        ("to_int", "0x4d2", 1234),
        ("to_str", "1234", "1234"),
        ("to_str", 1234, "1234"),
        ("to_epoch_ms", "2024-03-29T12:57:56.926Z", 1711717076926),
        ("to_epoch_ms", "2024-03-29T12:57:56.92Z", 1711717076920),
        ("to_epoch_ms", "2024-03-29T12:57:56.9Z", 1711717076900),
        ("to_epoch_ms", "2024-03-29T12:57:56Z", 1711717076000),
        ("lowercase", "WORKSTATION5.example.com", "workstation5.example.com"),
    ]
)
def test_run_transformer(transform, value, expected):
    assert run_transformer(transform, value) == expected


def test_run_series_basename():
    data = pd.Series([r"C:\Windows\System32\cmd.exe", r"C:\TMP"])
    result = list(run_transformer_on_series("basename", data))
    assert result == ["cmd.exe", "TMP"]



"""Kestrel Data Model Map value transformers"""

from datetime import datetime, timezone
from typing import Callable

from kestrel.exceptions import InvalidTransformerInMapping
from kestrel.mapping.path import Path
from pandas import Series

# Dict of "registered" transformers
_transformers = {}


def transformer(func: Callable) -> Callable:
    """A decorator for registering a transformer"""
    _transformers[func.__name__] = func
    return func


@transformer
def to_epoch_ms(value: str) -> int:
    """Convert a time value to milliseconds since the epoch"""
    if "." in value:
        time_pattern = "%Y-%m-%dT%H:%M:%S.%fZ"
    else:
        time_pattern = "%Y-%m-%dT%H:%M:%SZ"
    dt = datetime.strptime(value, time_pattern).replace(tzinfo=timezone.utc)
    return int(dt.timestamp() * 1000)


@transformer
def dirname(path: str) -> str:
    """Get the directory part of `path`"""
    return Path(path).dirname()


@transformer
def basename(path: str) -> str:
    """Get the filename part of `path`"""
    return Path(path).basename()


@transformer
def startswith(value: str) -> str:
    return f"{value}%"


@transformer
def winpath_startswith(value: str) -> str:
    return f"{value}\\%"


@transformer
def posixpath_startswith(value: str) -> str:
    return f"{value}/%"


@transformer
def endswith(value: str) -> str:
    return f"%{value}"


@transformer
def winpath_endswith(value: str) -> str:
    return f"%\\{value}"


@transformer
def posixpath_endswith(value: str) -> str:
    return f"%/{value}"


@transformer
def to_int(value) -> int:
    """Ensure `value` is an int"""
    try:
        return int(value)
    except ValueError:
        # Maybe it's a hexadecimal string?
        try:
            return int(value, 16)
        except:
            return -1


@transformer
def to_str(value) -> str:
    """Ensure `value` is a str"""
    return str(value)


@transformer
def lowercase(value: str) -> str:
    """Ensure `value` is all lowercase"""
    return value.lower()


@transformer
def ip_version_to_network_layer(value: int) -> str:
    if value == 4:
        return "ipv4"
    elif value == 6:
        return "ipv6"
    elif value == 99:
        return "other"
    return "unknown"


@transformer
def network_layer_to_ip_version(val: str) -> int:
    value = val.lower()
    if value == "ipv4":
        return 4
    elif value == "ipv6":
        return 6
    elif value == "other":
        return 99
    return 0


def run_transformer(transformer_name: str, value):
    """Run the registered transformer with name `transformer_name` on `value`"""
    func = _transformers.get(transformer_name)
    if func:
        result = func(value)
    else:
        raise InvalidTransformerInMapping(transformer_name)
    return result


def run_transformer_on_series(transformer_name: str, value: Series):
    """Run the registered transformer with name `transformer_name` on `value`"""
    func = _transformers.get(transformer_name)
    if func:
        result = value.apply(func)
    else:
        raise InvalidTransformerInMapping(transformer_name)
    return result



import logging
from collections import defaultdict
from functools import reduce
from typing import Callable, List, Optional, Union

import sqlalchemy
from kestrel.exceptions import (
    EntityNotFound,
    InvalidAttributes,
    InvalidMappingWithMultipleIdentifierFields,
    InvalidProjectEntityFromEntity,
    SourceSchemaNotFound,
)
from kestrel.ir.filter import (
    AbsoluteTrue,
    BoolExp,
    ExpOp,
    FBasicComparison,
    ListOp,
    MultiComp,
    NumCompOp,
    RefComparison,
    StrComparison,
    StrCompOp,
)
from kestrel.ir.instructions import (
    Filter,
    Information,
    Instruction,
    Limit,
    Offset,
    ProjectAttrs,
    ProjectEntity,
    Sort,
    SortDirection,
)
from kestrel.mapping.data_model import (
    translate_comparison_to_native,
    translate_projection_to_native,
)
from kestrel.mapping.utils import get_type_from_projection
from pandas import DataFrame
from pandas.io.sql import SQLTable, pandasSQL_builder
from sqlalchemy import and_, asc, column, desc, or_, select, tuple_
from sqlalchemy.engine import Compiled, Connection, default
from sqlalchemy.sql.elements import BinaryExpression, BooleanClauseList
from sqlalchemy.sql.expression import CTE, ColumnElement, ColumnOperators
from sqlalchemy.sql.functions import coalesce
from sqlalchemy.sql.selectable import Select
from typeguard import typechecked

_logger = logging.getLogger(__name__)

# SQLAlchemy comparison operator functions
comp2func = {
    NumCompOp.EQ: ColumnOperators.__eq__,
    NumCompOp.NEQ: ColumnOperators.__ne__,
    NumCompOp.LT: ColumnOperators.__lt__,
    NumCompOp.LE: ColumnOperators.__le__,
    NumCompOp.GT: ColumnOperators.__gt__,
    NumCompOp.GE: ColumnOperators.__ge__,
    StrCompOp.EQ: ColumnOperators.__eq__,
    StrCompOp.NEQ: ColumnOperators.__ne__,
    StrCompOp.LIKE: ColumnOperators.like,
    StrCompOp.NLIKE: ColumnOperators.not_like,
    StrCompOp.MATCHES: ColumnOperators.regexp_match,
    StrCompOp.NMATCHES: ColumnOperators.regexp_match,  # Caller must negate
    ListOp.IN: ColumnOperators.in_,
    ListOp.NIN: ColumnOperators.not_in,
}


@typechecked
class _TemporaryTable(SQLTable):
    def _execute_create(self):
        self.table = self.table.to_metadata(self.pd_sql.meta)
        self.table._prefixes.append("TEMPORARY")
        with self.pd_sql.run_transaction():
            self.table.create(bind=self.pd_sql.con)

    def create(self) -> None:
        # SQLite actually use the same temp DB everytime
        # even after connection closed and opened again
        # so we need to drop the previous temp table
        #
        # override the superclass code here
        # since we need to drop a table with a different schema
        # which is not supported by the superclass method
        if self.exists():
            with self.pd_sql.run_transaction():
                self.pd_sql.get_table(self.name).drop(bind=self.pd_sql.con)
                self.pd_sql.meta.clear()
        self._execute_create()


@typechecked
def ingest_dataframe_to_temp_table(conn: Connection, df: DataFrame, table_name: str):
    with pandasSQL_builder(conn) as pandas_engine:
        # no need to put if_exists="replace"
        # since our customized .create() only has this logic
        table = _TemporaryTable(table_name, pandas_engine, frame=df, index=False)
        table.create()
        df.to_sql(table_name, con=conn, if_exists="append", index=False)


def get_proj_cols(pairs):
    # Check for alias "collisions"
    alias_map = defaultdict(list)
    for i, j in pairs:
        alias_map[j].append(i)
    result = []
    coalesced = set()
    for i, j in pairs:
        if j in coalesced:
            continue  # Already handled
        native_fields = alias_map[j]
        if len(native_fields) > 1:
            # Multiple native fields are mapped to same alias - coalesce
            result.append(
                coalesce(*(sqlalchemy.column(field) for field in native_fields)).label(
                    j
                )
            )
            coalesced.add(j)
        else:
            result.append(sqlalchemy.column(i).label(j))
    return result


@typechecked
class SqlTranslator:
    def __init__(
        self,
        dialect: default.DefaultDialect,
        from_obj: Union[CTE, str],
        from_obj_schema: Optional[List[str]],  # Entity CTE does not require this
        from_obj_projection_base_field: Optional[str],
        ocsf_to_native_mapping: Optional[dict],  # CTE does not require this
        timefmt: Optional[Callable],  # CTE does not have time
        timestamp: Optional[str],  # CTE does not have time
    ):
        # Specify the schema if not Entity CTE
        # Event CTE and raw datasource need this for ProjectEntity
        self.source_schema = from_obj_schema

        # schema after projection
        # pass through the schema if no add_ProjectEntity()
        self.projected_schema = from_obj_schema

        # Store the mapping for translation from OCSF to native
        self.data_mapping = ocsf_to_native_mapping

        # SQLAlchemy Dialect object (e.g. from sqlalchemy.dialects import sqlite; sqlite.dialect())
        self.dialect = dialect

        # inherit projection_base_field from subquery
        self.projection_base_field = from_obj_projection_base_field

        # Time formatting function for datasource
        self.timefmt = timefmt

        # Primary timestamp field in target table
        self.timestamp = timestamp

        if isinstance(from_obj, CTE):
            from_clause = from_obj
            self.is_subquery = True
        else:
            from_clause = sqlalchemy.table(from_obj)
            self.is_subquery = False

        # SQLAlchemy statement object
        # Auto-dedup by default
        self.query: Select = select("*").select_from(from_clause).distinct()

    @typechecked
    def _map_identifier_field(self, field) -> ColumnElement:
        if self.data_mapping:
            comps = translate_comparison_to_native(self.data_mapping, field, "", None)
            if len(comps) > 1:
                raise InvalidMappingWithMultipleIdentifierFields(comps)
            else:
                col = column(comps[0][0])
        else:
            col = column(field)
        return col

    @typechecked
    def _render_comp(self, comp: FBasicComparison) -> BinaryExpression:
        if isinstance(comp, RefComparison):
            # most FBasicComparison has .field; RefComparison has .fields
            # col: ColumnElement
            if len(comp.fields) == 1:
                col = self._map_identifier_field(comp.fields[0])
            else:
                col = tuple_(
                    *[self._map_identifier_field(field) for field in comp.fields]
                )
            rendered_comp = comp2func[comp.op](col, comp.value)
        elif self.data_mapping:
            comps = translate_comparison_to_native(
                self.data_mapping, comp.field, comp.op, comp.value
            )
            if self.is_subquery:
                # do not translate field
                # only translate value
                comps = [(comp.field, op, value) for (_, op, value) in comps]

            translated_comps = (
                (
                    ~comp2func[op](column(field), value)
                    if op == StrCompOp.NMATCHES
                    else comp2func[op](column(field), value)
                )
                for field, op, value in comps
            )
            rendered_comp = reduce(or_, translated_comps)
        else:  # no translation
            rendered_comp = (
                ~comp2func[comp.op](column(comp.field), comp.value)
                if comp.op == StrCompOp.NMATCHES
                else comp2func[comp.op](column(comp.field), comp.value)
            )
        return rendered_comp

    @typechecked
    def _render_multi_comp(
        self, comps: MultiComp
    ) -> Union[BooleanClauseList, BinaryExpression]:
        op = and_ if comps.op == ExpOp.AND else or_
        binary_expressions = list(map(self._render_comp, comps.comps))

        # dedup using the SQLAlchemy's .compare() method; __eq__ does not work
        final_result = []
        for be in binary_expressions:
            for ue in final_result:
                if ue.compare(be):
                    break
            else:
                final_result.append(be)

        return reduce(op, final_result)

    @typechecked
    def _render_true(self) -> ColumnElement:
        return sqlalchemy.true()

    @typechecked
    def _render_exp(self, exp: BoolExp) -> ColumnElement:
        if isinstance(exp.lhs, AbsoluteTrue):
            lhs = self._render_true()
        elif isinstance(exp.lhs, BoolExp):
            lhs = self._render_exp(exp.lhs)
        elif isinstance(exp.lhs, MultiComp):
            lhs = self._render_multi_comp(exp.lhs)
        else:
            lhs = self._render_comp(exp.lhs)

        if isinstance(exp.rhs, AbsoluteTrue):
            rhs = self._render_true()
        elif isinstance(exp.rhs, BoolExp):
            rhs = self._render_exp(exp.rhs)
        elif isinstance(exp.rhs, MultiComp):
            rhs = self._render_multi_comp(exp.rhs)
        else:
            rhs = self._render_comp(exp.rhs)

        return and_(lhs, rhs) if exp.op == ExpOp.AND else or_(lhs, rhs)

    @typechecked
    def filter_to_selection(self, filt: Filter) -> ColumnElement:
        if filt.timerange.start:
            # Convert the timerange to the appropriate pair of comparisons
            start_comp = StrComparison(
                self.timestamp, ">=", self.timefmt(filt.timerange.start)
            )
            stop_comp = StrComparison(
                self.timestamp, "<", self.timefmt(filt.timerange.stop)
            )
            # AND them together
            time_exp = BoolExp(start_comp, ExpOp.AND, stop_comp)
            # AND that with any existing filter expression
            exp = BoolExp(filt.exp, ExpOp.AND, time_exp)
        else:
            exp = filt.exp
        if isinstance(exp, AbsoluteTrue):
            selection = self._render_true()
        elif isinstance(exp, BoolExp):
            selection = self._render_exp(exp)
        elif isinstance(exp, MultiComp):
            selection = self._render_multi_comp(exp)
        else:
            selection = self._render_comp(exp)
        return selection

    def add_Filter(self, filt: Filter) -> None:
        selection = self.filter_to_selection(filt)
        self.query = self.query.where(selection)

    def add_ProjectAttrs(self, proj: ProjectAttrs) -> None:
        if not self.source_schema:
            raise SourceSchemaNotFound(self.result_w_literal_binds())
        else:
            if self.source_schema != ["*"]:
                invalid_attrs = set(proj.attrs) - set(self.source_schema)
                if invalid_attrs:
                    raise InvalidAttributes(list(invalid_attrs))
            cols = [column(col) for col in proj.attrs]
            self.query = self.query.with_only_columns(*cols)

    def add_ProjectEntity(self, proj: ProjectEntity) -> None:
        if self.projection_base_field and self.projection_base_field != "event":
            raise InvalidProjectEntityFromEntity(proj, self.projection_base_field)
        else:
            self.projection_base_field = proj.ocsf_field

        if proj.ocsf_field == "event":  # project to event
            if self.data_mapping and not self.is_subquery:
                if not self.source_schema:
                    raise SourceSchemaNotFound(self.result_w_literal_binds())
                else:
                    pairs = translate_projection_to_native(
                        self.data_mapping, None, None, self.source_schema
                    )
            else:
                pairs = None
                _logger.debug("no data mapping, no translation for projection (event)")

        else:  # project to entity
            if not self.source_schema:
                raise SourceSchemaNotFound(self.result_w_literal_binds())

            if self.data_mapping and not self.is_subquery:
                pairs = translate_projection_to_native(
                    self.data_mapping, proj.ocsf_field, None, self.source_schema
                )
            else:
                prefix = proj.ocsf_field + "."
                pairs = [
                    (col, col[len(prefix) :])
                    for col in self.source_schema
                    if col.startswith(prefix)
                ]
            if not pairs and self.source_schema != ["*"]:
                # self.source_schema == ["*"] is used in virtual cache (EXPLAIN)
                entity_type = get_type_from_projection(proj.ocsf_field)
                raise EntityNotFound(
                    f"No '{entity_type}' found at '{proj.ocsf_field}.*' against the data source."
                )

        if pairs:
            self.projected_schema = [ocsf_field for _, ocsf_field in pairs]
            _logger.debug(f"column projection pairs: {pairs}")
            cols = get_proj_cols(pairs)
            self.query = self.query.with_only_columns(*cols)

    def add_Limit(self, lim: Limit) -> None:
        self.query = self.query.limit(lim.num)

    def add_Information(self, ins: Information) -> None:
        self.query = self.query.limit(1)

    def add_Offset(self, offset: Offset) -> None:
        self.query = self.query.offset(offset.num)

    def add_Sort(self, sort: Sort) -> None:
        col = column(sort.attribute)
        order = asc(col) if sort.direction == SortDirection.ASC else desc(col)
        self.query = self.query.order_by(order)

    def add_instruction(self, i: Instruction) -> None:
        inst_name = i.instruction
        method_name = f"add_{inst_name}"
        method = getattr(self, method_name)
        if not method:
            raise NotImplementedError(f"SqlTranslator.{method_name}")
        method(i)

    def result(self) -> Compiled:
        return self.query.compile(dialect=self.dialect)

    def result_w_literal_binds(self) -> Compiled:
        # full SQL query with literal binds showing, i.e., IN [99, 51], not IN [?, ?]
        # this is for debug display, not used by an sqlalchemy driver to execute
        return self.query.compile(
            dialect=self.dialect, compile_kwargs={"literal_binds": True}
        )



import yaml
from kestrel.utils import list_folder_files

MAPPING_MODULE = "kestrel.mapping"

# cache type mapping
type_mapping = {}


def get_type_from_projection(proj) -> str:
    global type_mapping
    if not type_mapping:
        for f in list_folder_files(MAPPING_MODULE, "types", extension="yaml"):
            with open(f, "r") as fp:
                mapping_ind = yaml.safe_load(fp)
                type_mapping.update(mapping_ind)
    if proj in type_mapping:
        return type_mapping[proj]
    else:
        return "unkown_entity"



import logging
import re
from datetime import datetime
from typing import Callable, Iterable, List, Tuple

import lark
from kestrel.config.utils import get_all_relations
from kestrel.frontend.parser import get_keywords, parse_without_transform
from kestrel.interface import InterfaceManager
from kestrel.interface.base import AnalyticsInterface, DatasourceInterface
from kestrel.mapping.data_model import get_all_entity_types
from kestrel.utils import timefmt
from typeguard import typechecked

_logger = logging.getLogger(__name__)

ISO_TS_RE = re.compile(r"\d{4}(-\d{2}(-\d{2}(T\d{2}(:\d{2}(:\d{2}Z?)?)?)?)?)?")


@typechecked
def do_complete(
    code: str,
    cursor_pos: int,
    itf_manager: InterfaceManager,
    varnames: List[str],
) -> Iterable[str]:
    _logger.debug("auto_complete function starts...")

    # do not care code after cursor position in the current version
    line = code[:cursor_pos]
    _logger.debug(f"line to auto-complete: {line}")

    # if the last char is a space, `line_to_parse = line`
    # otherwise, exclude the last token in `line_to_parse` to prompt the expected token
    last_word_prefix, line_to_parse = _split_last_token(line)
    _logger.debug(f"last word prefix: {last_word_prefix}")
    _logger.debug(f"line to parse: {line_to_parse}")

    try:
        ast = parse_without_transform(line_to_parse)

    except lark.exceptions.UnexpectedCharacters as e:
        suggestions = ["% illegal char in huntflow %"]
        _logger.debug(f"illegal character in `line_to_parse`, err: {str(e)}")

    except lark.exceptions.UnexpectedEOF as e:
        suggestions = ["% EOF auto-complete internal error, report to developers %"]
        # https://github.com/lark-parser/lark/issues/791
        # Lark updates may break this, check if it is the case
        # no need to use KestrelInternalError; not to break huntflow execution
        _logger.debug(f"Lark with LALR should not give this error: {str(e)}")

    except lark.exceptions.UnexpectedToken as e:
        error_token = e.token
        expected_tokens = e.accepts or e.expected
        expected_values = []
        keywords = set(get_keywords(False))
        relations = get_all_relations()
        entity_types = get_all_entity_types()
        for token in expected_tokens:
            _logger.debug("token: %s", token)
            if token == "VARIABLE":
                expected_values.extend(varnames)
            elif token == "ISOTIMESTAMP":
                if last_word_prefix:
                    if last_word_prefix.startswith("t'"):
                        ts_prefix = last_word_prefix[2:]
                        ts_complete = _do_complete_timestamp(ts_prefix)
                        exp_value = "t'" + ts_complete + "'"
                    else:
                        exp_value = _do_complete_timestamp(last_word_prefix)
                else:
                    exp_value = timefmt(datetime.now())
                expected_values.append(exp_value)
            elif token == "DATASRC_SIMPLE":
                _logger.debug("auto-complete data source")
                expected_values.extend(
                    _do_complete_interface(
                        last_word_prefix,
                        itf_manager.schemes(DatasourceInterface),
                        itf_manager.list_datasources_from_scheme,
                    )
                )
            elif token == "ANALYTICS_SIMPLE":
                _logger.debug("auto-complete analytics")
                expected_values.extend(
                    _do_complete_interface(
                        last_word_prefix,
                        itf_manager.schemes(AnalyticsInterface),
                        itf_manager.list_datasources_from_scheme,
                    )
                )
            elif token == "ENTITY_TYPE":
                expected_values.extend(entity_types)
            elif token == "PROJECT_FIELD":  # not precise
                expected_values.extend(entity_types)
            elif token == "RELATION":
                expected_values.extend(relations)
            elif token == "REVERSED":
                expected_values.append("BY")
            elif token == "EQUAL":
                expected_values.append("=")
            elif token == "ATTRIBUTE":
                # TODO: attribute completion
                # https://github.com/opencybersecurityalliance/kestrel-lang/issues/79
                _logger.debug(f"TODO: ATTRIBUTE COMPLETION")
            elif token == "ENTITY_ATTRIBUTE_PATH":
                # TODO: attribute completion
                # https://github.com/opencybersecurityalliance/kestrel-lang/issues/79
                _logger.debug(f"TODO: ATTRIBUTE COMPLETION")
            elif token == "COMMA":
                expected_values.append(",")
            elif token in keywords:
                if last_word_prefix and last_word_prefix.islower():
                    token = token.lower()
                expected_values.append(token)
            else:
                # token not handled
                continue
        expected_values = sorted(expected_values)
        _logger.debug(f"expected values: {expected_values}")

        # turn `expected_values` into `suggestions`
        _p = last_word_prefix
        _e = expected_values
        suggestions = [t[len(_p) :] for t in _e if t.startswith(_p)] if _p else _e
        suggestions = [x for x in set(suggestions) if x]
        _logger.debug(f"suggestions: {suggestions}")

    else:
        suggestions = []

        # handle optional components
        if ast:
            if ast.children:
                stmt = ast.children[-1].children[0]
                cmd = (
                    stmt.children[1].data.value
                    if stmt.data.value == "assignment"
                    else stmt.data.value
                )
                if cmd == "disp":
                    for clause in ("attr_clause", "limit_clause", "offset_clause"):
                        if not list(stmt.find_data(clause)):
                            suggestions.append("ATTR")
                elif cmd in ("expression", "find") and not list(
                    stmt.find_data("where_clause")
                ):
                    suggestions.append("WHERE")
                elif cmd in ("get", "find") and not list(stmt.find_data("timerange")):
                    suggestions.append("START")
                elif cmd == "apply" and not list(stmt.find_data("args")):
                    suggestions.append("WITH")
            else:
                suggestions = ["DISP", "APPLY", "EXPLAIN", "INFO", "SAVE", "DESCRIBE"]

        suggestions = [x for x in set(suggestions) if x]
        _p = last_word_prefix
        suggestions = (
            [t[len(_p) :] for t in suggestions if t.startswith(_p)]
            if _p
            else suggestions
        )
        _logger.debug(f"suggestions from optional components: {suggestions}")

    return suggestions


@typechecked
def _end_with_blank_or_comma(s: str) -> bool:
    return s[-1] in [" ", "\t", "\n", "\r", "\f", "\v", ","] if s else True


@typechecked
def _split_last_token(s: str) -> Tuple[str, str]:
    last = ""
    if not _end_with_blank_or_comma(s):
        while not _end_with_blank_or_comma(s):
            last = s[-1] + last
            s = s[:-1]
    return last, s


@typechecked
def _do_complete_timestamp(ts_prefix: str) -> str:
    valid_ts_formats = [
        "%Y",
        "%Y-%m",
        "%Y-%m-%d",
        "%Y-%m-%dT%H",
        "%Y-%m-%dT%H:%M",
        "%Y-%m-%dT%H:%M:%S",
    ]
    matched = ISO_TS_RE.match(ts_prefix)
    if matched:
        for ts_format in valid_ts_formats:
            _logger.debug(f"Match timestamp {ts_prefix} with format {ts_format}")
            try:
                ts = datetime.strptime(matched.group(), ts_format)
            except:
                _logger.debug(f"Timestamp match failed")
            else:
                ts_complete = timefmt(ts)
                _logger.debug(f"Timestamp completed: {ts_complete}")
                break
        else:
            ts_complete = "% TS auto-complete internal error, report to developers %"
            # no need to use KestrelInternalError; not to break huntflow execution
            _logger.debug(
                f"TS auto-complete internal error: `valid_ts_formats` is incomplete"
            )
    else:
        ts_complete = "% illegal ISO 8601 timestamp prefix %"
        _logger.debug(f"illegal ISO 8601 timestamp prefix: {ts_prefix}")
    return ts_complete


@typechecked
def _do_complete_interface(
    last_word_prefix: str,
    schemes: Iterable[str],
    list_names_from_scheme: Callable,
) -> Iterable[str]:
    if last_word_prefix and "://" in last_word_prefix:
        scheme, _ = last_word_prefix.split("://")
        if scheme in schemes:
            names = list_names_from_scheme(scheme)
            paths = [scheme + "://" + name for name in names]
            _logger.debug(f"auto-complete interface {scheme}: {paths}")
            expected_values = paths
    else:
        expected_values = [scheme + "://" for scheme in schemes]
    return expected_values
